# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import uuid

import pytest

from karabo.middlelayer import call, getDevice, updateDevice
from karabo.middlelayer.testing import AsyncDeviceContext
from karabo.native import Hash

from ..macro_server import MacroServer

TEST_MACROSERVER = "Karabo_TestMacroServer"

SYNC_CODE = """from karabo.middlelayer import *

class TestMacro(Macro):
    s = String()

    @Slot()
    def do(self):
        self.s = "sync"
"""

ASYNC_CODE = """from karabo.middlelayer import *

class TestMacro(Macro):
    s = String()

    @Slot()
    async def do(self):
        self.s = "async"
"""


async def init_macro(code, expected):
    uuid_ = str(uuid.uuid4())
    config = Hash("uuid", uuid_, "module", "test",
                  "code", code)
    h = Hash("classId", "MetaMacro", "configuration", config,
             "deviceId", f"bla-{uuid_}")
    assert h["configuration"] == config
    await call(TEST_MACROSERVER, "slotStartDevice", h)
    proxy = await getDevice(f"bla-{uuid_}")
    with proxy:
        await updateDevice(proxy)
        await proxy.do()
        assert proxy.s.value == expected
    await call(proxy.deviceId, "slotKillDevice")


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_macroserver():
    server = MacroServer(dict(serverId=TEST_MACROSERVER))
    async with AsyncDeviceContext(server=server):
        await init_macro(SYNC_CODE, "sync")
        await init_macro(ASYNC_CODE, "async")



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import json
import os
import time
from argparse import ArgumentParser
from asyncio import ensure_future, get_event_loop, sleep, wait_for
from copy import deepcopy

from tornado import web
from tornado.httpclient import AsyncHTTPClient
from tornado.platform.asyncio import AsyncIOMainLoop, to_asyncio_future

from .startkarabo import entrypoint
from .webserver import EMPTY_NETWORK_RESPONSE, EMPTY_RESPONSE

HEARTBEAT_PENALTY = 10  # the server is polled every 5 seconds see webserver.py
CHECK_INTERVAL = 20


class MainHandler(web.RequestHandler):
    def initialize(self, servers, client, last_updates, network):
        self.servers = servers
        self.network = network
        self.client = client
        self.last_updates = last_updates

    def get(self):
        self.render("aggregator_index.html", servers=self.servers)

    async def request_servers(self, hostname, port):
        khostname, *_ = hostname.split('.')  # the hostname without domain
        url = f"http://{hostname}:{port}"
        api_url = f"{url}/api/servers.json"
        server = self.servers.setdefault(
            khostname, dict(hostname=hostname, link=url))
        f = to_asyncio_future(self.client.fetch(api_url, method="GET"))
        r = await wait_for(f, timeout=2)
        assert r.code == 200
        h = json.loads(r.body)
        # filter servers that cannot be controlled.
        table = [server for server in h['servers']
                 if server['control_allowed']]
        table.sort(key=lambda x: x['karabo_name'])
        server["services"] = table

    async def request_network(self, hostname, port):
        khostname, *_ = hostname.split('.')  # the hostname without domain
        url = f"http://{hostname}:{port}"
        api_url = f"{url}/api/network.json"
        net = self.network.setdefault(
            khostname, dict(hostname=hostname))
        f = to_asyncio_future(self.client.fetch(api_url, method="GET"))
        r = await wait_for(f, timeout=2)
        assert r.code == 200
        h = json.loads(r.body)
        net["network"] = h["network"]

    async def post(self):
        try:
            port = self.get_argument("port")
            hostname = self.get_argument("hostname")
        except web.MissingArgumentError:
            raise web.HTTPError(
                status_code=400,  # bad request
                log_message='')
        ensure_future(self.request_servers(hostname, port))

        # New web servers also provide the version
        try:
            self.get_argument("version")
            ensure_future(self.request_network(hostname, port))
        except web.MissingArgumentError:
            khostname, *_ = hostname.split('.')
            net = self.network.setdefault(
                khostname, dict(hostname=hostname))
            net["error"] = "Not supported in current version: 1.0.0"
        self.last_updates[(hostname, port)] = time.time()
        self.write("OK")


class StatusHandler(web.RequestHandler):
    def initialize(self, servers, client, last_updates, network):
        self.servers = servers

    def get(self):
        response = deepcopy(EMPTY_RESPONSE)
        response['servers'] = self.servers
        response['success'] = True
        self.write(response)


class NetworkHandler(web.RequestHandler):
    def initialize(self, servers, client, last_updates, network):
        self.network = network

    def get(self):
        response = deepcopy(EMPTY_NETWORK_RESPONSE)
        response['network'] = self.network
        response['success'] = True
        self.write(response)


class HeartbeatChecker:
    def __init__(self, servers, client, last_updates, network):
        self.servers = servers
        self.last_updates = last_updates

    async def ensure_hosts_alive(self):
        while True:
            registered = deepcopy(self.last_updates)
            now = time.time()
            for key, last_update in registered.items():
                hostname, _ = key
                if last_update < now - HEARTBEAT_PENALTY:
                    # the hostname shown in karabo
                    khostname, *_ = hostname.split('.')
                    self.servers.pop(khostname, None)
                    self.last_updates.pop(key)
            await sleep(CHECK_INTERVAL)


@entrypoint
def run_webserver():
    """karabo-webaggregatorserver - start a web server to aggregate
    karabo webservers

      karabo-webaggregatorserver serverId=webserver [-h|--help]
      [--port portnumber ]

    --port
      port number the server listens to
    """

    parser = ArgumentParser()
    parser.add_argument('serverId')
    parser.add_argument('--port',
                        help='port number the server listens to',
                        default=8585)
    args = parser.parse_args()
    if not args.serverId.startswith('serverId='):
        parser.print_help()
        return
    servers = dict()
    if hasattr(AsyncIOMainLoop, "initialized"):
        if not AsyncIOMainLoop.initialized():
            AsyncIOMainLoop().install()

    server_dict = {'servers': servers, 'client': AsyncHTTPClient(),
                   'last_updates': {}, 'network': {}}
    heartbeat = HeartbeatChecker(**server_dict)

    app = web.Application([('/', MainHandler, server_dict),
                           ('/status.json', StatusHandler, server_dict),
                           ('/network.json', NetworkHandler, server_dict)],
                          template_path=os.path.join(
                              os.path.dirname(__file__), "templates"),
                          static_path=os.path.join(
                              os.path.dirname(__file__), "static"),)
    app.listen(args.port)
    ensure_future(heartbeat.ensure_hosts_alive())
    get_event_loop().run_forever()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
LIST_DOMAINS_QUERY = """
    xquery version "3.0";
    <collections>{{
    for $c in xmldb:get-child-collections("{}")
    return <item>{{$c}}</item>}}
    </collections>
    """
ROOT_COLLECTION = "krb_config"
ROOT_COLLECTION_TEST = "krb_test"
ROOT_COLLECTION_BACKUP = "krb_backup"

DATE_FORMAT = '%Y-%m-%d %H:%M:%S'



#!/usr/bin/python
# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
#
# Author: <burkhard.heisen@xfel.eu>
#

import datetime
import re
import time

import IPython
import pytz
import tzlocal
from dateutil import parser

from karabo._version import version as karaboVersion
from karabo.bound import (
    DeviceClient as BoundDeviceClient, Hash, TextSerializerHash)

# ip = IPython.core.ipapi.get()
ip = IPython.get_ipython()

# Create one instance (global singleton) of a DeviceClient
cpp_client = None


# Welcome
print("\n#### Karabo Device-Client (version:", karaboVersion, ") ####")
print("To start you need a DeviceClient object, e.g. type:\n")
print("  d = DeviceClient()\n")
print("Using this object you can remote control Karabo devices.")
print("You may query servers and devices and set/get properties or execute "
      "commands on them.")
print("Hint: use the TAB key for auto-completion.")


def distributed_autocomplete(func):
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except BaseException:
            print("Distributed auto-completion failed")
    return wrapper


# Auto complete function for methods like: f( deviceId, attribute )
@distributed_autocomplete
def auto_complete_full(self, event):
    if (re.match(r'.*\(\s*$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*$', event.line)):
        return ["\""]
    elif (re.match(r'.*\(\s*\"\w+$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*\"\w+$', event.line)):
        # These REs match correctly but completion does not work
        return ["\""]
    elif (re.match(r'.*\(\s*\"$', event.line)):
        dev = cpp_client.getDevices()
        if len(dev) > 0:
            return dev
        else:
            return ["NO_INSTANCES_AVAILABLE"]
    elif (re.match(r'.*\(.+,\s*\"$', event.line)):
        r = re.compile(r'\"(.*?)\"')
        m = r.search(event.line)
        if m:
            deviceId = m.group(1)
            return cpp_client.getProperties(deviceId)
        else:
            return [""]
    else:
        return [""]


@distributed_autocomplete
def auto_complete_devid(self, event):
    if (re.match(r'.*\(\s*$', event.line)):
        return ["\""]
    elif (re.match(r'.*\(\s*\"\w+$', event.line)):
        # These REs match correctly but completion does not work
        return ["\""]
    elif (re.match(r'.*\(\s*\"$', event.line)):
        dev = cpp_client.getDevices()
        if len(dev) > 0:
            return dev
        else:
            return ["NO_INSTANCES_AVAILABLE"]
    else:
        return [""]


@distributed_autocomplete
def auto_complete_serverid(self, event):
    if (re.match(r'.*\(\s*$', event.line)):
        return ["\""]
    elif (re.match(r'.*\(\s*\"\w+$', event.line)):
        # These REs match correctly but completion does not work
        return ["\""]
    elif (re.match(r'.*\(\s*\"$', event.line)):
        dev = cpp_client.getServers()
        if len(dev) > 0:
            return dev
        else:
            return ["NO_SERVERS_AVAILABLE"]
    else:
        return [""]


@distributed_autocomplete
def auto_complete_set(self, event):
    if (re.match(r'.*\(\s*$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*$', event.line)):
        return ["\""]
    elif (re.match(r'.*\(\s*\"\w+$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*\"\w+$', event.line)):
        # These REs match correctly but completion does not work
        return ["\""]
    elif (re.match(r'.*\(\s*\"$', event.line)):
        dev = cpp_client.getDevices()
        if len(dev) > 0:
            return dev
        else:
            return ["NO_INSTANCES_AVAILABLE"]
    elif (re.match(r'.*\(.+,\s*\"$', event.line)):
        r = re.compile(r'\"(.*?)\"')
        m = r.search(event.line)
        if m:
            deviceId = m.group(1)
            return cpp_client.getCurrentlySettableProperties(deviceId)
        else:
            return [""]
    else:
        return [""]


@distributed_autocomplete
def auto_complete_execute(self, event):
    if (re.match(r'.*\(\s*$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*$', event.line)):
        return ["\""]
    elif (re.match(r'.*\(\s*\"\w+$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*\"\w+$', event.line)):
        # These REs match correctly but completion does not work
        return ["\""]
    elif (re.match(r'.*\(\s*\"$', event.line)):
        dev = cpp_client.getDevices()
        if len(dev) > 0:
            return dev
        else:
            return ["NO_INSTANCES_AVAILABLE"]
    elif (re.match(r'.*\(.+,\s*\"$', event.line)):
        r = re.compile(r'\"(.*?)\"')
        m = r.search(event.line)
        if m:
            deviceId = m.group(1)
            return cpp_client.getCurrentlyExecutableCommands(deviceId)
        else:
            return [""]
    else:
        return [""]


@distributed_autocomplete
def auto_complete_instantiate(self, event):
    if (re.match(r'.*\(\s*$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*$', event.line)):
        return ["\""]
    elif (re.match(r'.*\(\s*\"\w+$', event.line)
            or re.match(r'.*\(\s*\"\w+\",\s*\"\w+$', event.line)):
        # These REs match correctly but completion does not work
        return ["\""]
    elif (re.match(r'.*\(\s*\"$', event.line)):
        dev = cpp_client.getServers()
        if len(dev) > 0:
            return dev
        else:
            return ["NO_INSTANCES_AVAILABLE"]
    elif (re.match(r'.*\(.+\,\s*$', event.line)):
        return ["\""]
    elif (re.match(r'.*\(.+,\s*\"$', event.line)):
        r = re.compile(r'\"(.*?)\"')
        m = r.search(event.line)
        if m:
            serverId = m.group(1)
            return cpp_client.getClasses(serverId)
        else:
            return [""]
    elif (re.match(r'.*\(.+,.+,\s*$', event.line)
            or re.match(r'.*\(.+\,.+\,\s*\".*$', event.line)):
        return ["\""]
    else:
        return [""]


# Register hooks in IPython
if ip is not None:
    ip.set_hook('complete_command', auto_complete_full,
                re_key='.*getFromPast')
    ip.set_hook('complete_command', auto_complete_full,
                re_key='.*getPropertyHistory')
    ip.set_hook('complete_command', auto_complete_full,
                re_key='.*registerPropertyMonitor')
    ip.set_hook('complete_command', auto_complete_full,
                re_key='.*unregisterPropertyMonitor')
    ip.set_hook('complete_command', auto_complete_full,
                re_key='.*help')

    ip.set_hook('complete_command', auto_complete_serverid,
                re_key='.*shutdownServer')
    ip.set_hook('complete_command', auto_complete_serverid,
                re_key='.*getClasses')
    ip.set_hook('complete_command', auto_complete_serverid,
                re_key='.*getDevices')

    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*shutdownDevice')
    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*setHash')
    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*setPrio')
    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*lock')
    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*registerDeviceMonitor')
    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*unregisterDeviceMonitor')
    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*getDeviceHistory')
    ip.set_hook('complete_command', auto_complete_devid,
                re_key='.*getConfigurationFromPast')

    ip.set_hook('complete_command', auto_complete_set,
                re_key='.*setNoWait')
    ip.set_hook('complete_command', auto_complete_execute,
                re_key='.*execute')

    ip.set_hook('complete_command', auto_complete_instantiate,
                re_key='.*instantiate')
    ip.set_hook('complete_command', auto_complete_instantiate,
                re_key='.*getClassSchema')

    # Short keys go to end
    ip.set_hook('complete_command', auto_complete_full,
                re_key='.*get')
    ip.set_hook('complete_command', auto_complete_set,
                re_key='.*set')


class DeviceClient:
    """DeviceClient to remotely control Karabo

    The DeviceClient allows to remotely control a Karabo installation.
    A Karabo installation comprises all distributed end-points (servers,
    devices and clients), which talk to the same central message-broker as
    defined by its host, port and topic. The DeviceClient establishes a direct
    connection to the broker. You may specify which broker and topic should be
    used via the environment variables

      KARABO_BROKER       (default: amqp://xfel:karabo@exfl-broker-1:5672)
      KARABO_BROKER_TOPIC (default: $USER)

    where defaults stated above are given if the environment variable is not
    set.
    """

    def __init__(self):
        global cpp_client
        if cpp_client is None:
            cpp_client = BoundDeviceClient()
        self.__client = cpp_client

        self.values = dict()

    def instantiate(self, serverId, classId, deviceId, config=Hash(),
                    timeout=None):
        """Instantiate (and configure) a device on a running server.

        NOTE: This call is synchronous (blocking)

        Args:
        :param serverId: The serverId of the server on which the device should
                        be started.
        :param classId:  The classId of the device (corresponding plugin must
                        be loaded on the server)
        :param deviceId: The future name of the device in the Karabo
                        installation (will fail if not unique)
        :param config:   The initial configuration of the device (optional if
                        all parameters of the device are optional)
        :param timeout : Timeout in seconds until this function will be forced
                        to return
        :returns Tuple with (True, <deviceId>) in case of success or tuple with
                        (False, <errorMessage>) in case of failure
        """
        # This is hacked here and should be added to c++
        config.set("deviceId", deviceId)
        if timeout is None:
            return self.__client.instantiate(serverId, classId, config)
        return self.__client.instantiate(serverId, classId, config, timeout)

    def _instantiate(self, serverId, classId, deviceId, data, timeout=None):
        # serializer = BinarySerializerHash.create("Bin") # this does not work
        serializer = TextSerializerHash.create("Xml")
        self.instantiateNoWait(serverId, classId, deviceId,
                               serializer.load(data))

    def instantiateNoWait(self, serverId, classId, deviceId, config=Hash()):
        """Instantiate (and configure) a device on a running server.

        NOTE: This call is asynchronous (non-blocking)

        Args:
        :param serverId: The serverId of the server on which the device should
                        be started.
        :param classId:  The classId of the device (corresponding plugin must
                        be loaded on the server)
        :param deviceId: The future name of the device in the Karabo
                        installation (will fail if not unique)
        :param config:   The initial configuration of the device (optional if
                        all parameters of the device are optional)
        """
        # This is hacked here and should be added to c++
        config.set("deviceId", deviceId)
        self.__client.instantiateNoWait(serverId, classId, config)

    def shutdownDevice(self, deviceId, timeout=None):
        """Shuts down a device.

        NOTE: This call is synchronous (blocking)

        Args:
        :param deviceId: The deviceId of the device to be destructed.
        :param timeout : Timeout in seconds until this function will be forced
                        to return
        :returns Tuple with (True, <deviceId>) in case of success or tuple with
                        (False, <errorMessage>) in case of failure
        """
        if timeout is None:
            return self.__client.killDevice(deviceId)
        return self.__client.killDevice(deviceId, timeout)

    def shutdownDeviceNoWait(self, deviceId):
        """Shuts down a device.

        NOTE: This call is asynchronous (non-blocking)

        Args:
        :param deviceId: The deviceId of the device to be destructed.
        """
        self.__client.killDeviceNoWait(deviceId)

    def shutdownServer(self, serverId, timeout=None):
        """Shuts down a server.

        NOTE: This call is synchronous (blocking)

        Args:
        :param serverId: The serverId of the server to be destructed.
        :param timeout : Timeout in seconds until this function will be forced
                        to return
        :returns Tuple with (True, <serverId>) in case of success or tuple with
                (False, <errorMessage>) in case of failure
        """
        if timeout is None:
            return self.__client.killServer(serverId)
        return self.__client.killServer(serverId, timeout)

    def shutdownServerNoWait(self, serverId):
        """Shuts down a server.

        NOTE: This call is asynchronous (non-blocking)

        Args:
        :param serverId: The serverId of the server to be destructed.
        """
        self.__client.killServerNoWait(serverId)

    def getServers(self):
        """Returns a list of currently running servers."""
        return self.__client.getServers()

    def getDevices(self, serverId=None):
        """Returns a list of currently running devices.

        Args:
        :param serverId: Optionally only the running devices of a given server
                        can be listed.
        """
        if serverId is None:
            return self.__client.getDevices()
        return self.__client.getDevices(serverId)

    def getClasses(self, serverId):
        """Returns a list of available device classes (plugins) on a server

        Args:
        :param serverId: The server of whose plugins should be listed.
        """
        return self.__client.getClasses(serverId)

    def help(self, instanceId, parameter=None):
        """Provides help on device instance or parameter

        This function provides help on a full instance or a specific
        parameter of an instance. Instance is also called the name or address
        of the specific device.
        """
        if parameter is None:
            self.__client.getDeviceSchema(instanceId).help()
        else:
            self.__client.getDeviceSchema(instanceId).help(parameter)

    def get(self, instanceId, propertyName=None):
        """Retrieves attribute values from device

        It only retrieves attributes marked for DAQ, throws an exception
        otherwise.
        """
        if propertyName is None:
            return self.__client.get(instanceId)
        return self.__client.get(instanceId, propertyName)

    def getFromPast(self, deviceId, propertyName, t0, t1=None,
                    maxNumData=10000):
        """Deprecated, use getPropertyHistory instead"""
        return self.getPropertyHistory(deviceId, propertyName, t0, t1=t1,
                                       maxNumData=maxNumData)

    def getPropertyHistory(self, deviceId, propertyName, t0, t1=None,
                           maxNumData=10000):
        """Get the history of device properties

        With this function one can get all values of a property in a given
        timespan::

        getPropertyHistory(deviceId, propertyName, "2015-12-01", "2015-12-02")

        returns a list of Hashes, which contain all changes of *propertyName*
        between the two given dates. Each Hash has a node with key 'v'. Its
        value is the one of the property at the time defined by the attributes
        'sec' and 'frac' which holds the seconds and attoseconds, respectively,
        since 1970-01-01 UTC. The attribute 'tid' holds the train ID.

        The dates of the timespan are parsed using
        :func:`dateutil.parser.parse`, allowing many ways to write the date.
        The most precise way is to write "2015-12-01T15:32:12 UTC", but you may
        omit any part, like "10:32", only giving the time, where we assume
        the current day.  Unless specified otherwise, your local timezone is
        assumed.

        Another parameter, *maxNumData*, may be given, which gives the maximum
        number of data points to be returned. The returned data will be
        reduced appropriately to still span the full timespan. *maxNumData=0*
        means no reduction - note that the history request might timeout.
        """
        utc_t0 = self._fromTimeStringToUtcString(t0)
        utc_t1 = self._fromTimeStringToUtcString(
            t1) if t1 is not None else datetime.datetime.now().isoformat()

        return self.__client.getFromPast(deviceId, propertyName, utc_t0,
                                         utc_t1, maxNumData)

    def getDeviceHistory(self, deviceId, timepoint):
        """Get configuration of a device at a given time point

        Similar to 'getConfigurationFromPast' of the C++ DeviceClient.
        Concerning the format of the timepoint, see getPropertyHistory.
        """
        return self.getConfigurationFromPast(deviceId, timepoint)

    def getConfigurationFromPast(self, deviceId, timepoint):
        """Same as getDeviceHistory

        Kept for interface similarity with the C++ DeviceClient.
        """
        utc_timepoint = self._fromTimeStringToUtcString(timepoint)
        return self.__client.getConfigurationFromPast(deviceId, utc_timepoint)

    def getSystemInformation(self):
        return self.__client.getSystemInformation()

    def getSystemTopology(self):
        return self.__client.getSystemTopology()

    def getClassSchema(self, serverId, classId):
        return self.__client.getClassSchema(serverId, classId)

    def getDeviceSchema(self, deviceId):
        return self.__client.getDeviceSchema(deviceId)

    def getDeviceSchemaNoWait(self, deviceId):
        return self.__client.getDeviceSchemaNoWait(deviceId)

    def registerSchemaUpdatedMonitor(self, callbackFunction):
        """Registers an async call-back on schema update

        This function can be used to register an asynchronous call-back on
        schema update from the distributed system.

        Args:
        :param callbackFunction: the call-back function to be registered.
                    It must have the following signature: f(str, Schema)

        Note:
        Currently, registering only a schema update monitor with an instance
        of a DeviceClient is not enough to have the registered call-back
        activated.
        A workaround for this is to also register a property monitor with the
        same instance of DeviceClient that has been used to register the schema
        update monitor.

        Example:

        def onSchemaUpdate(deviceId, schema):
            print("{}: {}".format(deviceId, schema))

        def onPropertyChange(deviceId, key, value, timeStamp):
            # In this example the property monitor is empty.
            pass

        c = DeviceClient()
        c.registerSchemaUpdatedMonitor(onSchemaUpdate)
        c.registerPropertyMonitor("Test_Device_0", "result", onPropertyChange)
        """
        return self.__client.registerSchemaUpdatedMonitor(callbackFunction)

    def registerDeviceMonitor(self, instanceId, callbackFunction,
                              userData=None):
        """Registers an async call-back on change of device property

        This function can be used to register an asynchronous call-back on
        change of any device property. The call-back function must have the
        following signature: f(str, Hash)
        arg1: deviceId
        arg2: currently changed part of the configuration

        Example:

        def onDeviceChange(deviceId, config):
            print deviceId, ":", config

        c = DeviceClient();
        c.registerDeviceMonitor("Test_MyDevice_0", onDeviceChange)
        """
        if userData is None:
            return self.__client.registerDeviceMonitor(
                instanceId, callbackFunction)
        return self.__client.registerDeviceMonitor(
            instanceId, callbackFunction, userData)

    def unregisterDeviceMonitor(self, instanceId):
        self.__client.unregisterDeviceMonitor(instanceId)

    def registerChannelMonitor(self, channelName, dataHandler=None,
                               inputChannelCfg=None, eosHandler=None,
                               inputHandler=None, statusTracker=None):
        """Register an asynchronous call-back to monitor defined output channel

        Internally, an InputChannel is created and configured.
        :param channelName identifies the channel as a concatenation of the id
                            of its device, a colon (:) and the name of the
                            output channel (e.g. A/COOL/DEVICE:output)
        :param dataHandler called when data arrives, arguments are data (Hash)
                            and meta data (Hash/MetaData)
        :param inputChannelCfg configures the channel via
                        InputChannel.create(..) use default except you know
                        what your are doing.
                        For experts: "connectedOutputChannels" will be
                                        overwritten.
                                      "onSlowness" default is overwritten
                                        to "drop"
        :param eosHandler called on end of stream,
                        argument is the InputChannel
        :param inputHandler called when data arrives,
                        argument is the InputChannel
        :param statusTracker called with a 'ConnectionStatus' as argument when
                             the connection status of the underlying
                             InputChannel changes.

        :returns False if channel is already registered

        Example:

        def handler(data, meta):
            print(data.getPaths())
        def tracker(status):
            if status is ConnectionStatus.CONNECTED:
                print("Ready for data")
            elif status is ConnectionStatus.DISCONNECTED:
                print("Lost data connection")
        c = DeviceClient();
        c.registerChannelMonitor("DEV/ID/1:output", handler,
                                 statusTracker=tracker)
        """
        if inputChannelCfg is None:
            inputChannelCfg = Hash()
        return self.__client.registerChannelMonitor(channelName, dataHandler,
                                                    inputChannelCfg,
                                                    eosHandler, inputHandler,
                                                    statusTracker)

    def unregisterChannelMonitor(self, channelName):
        """Unregister monitoring of output channel

        :param channelName identifies the channel as a concatenation of the id
                            of its devices, a colon (:) and the name of the
                            output channel (e.g. A/COOL/DEVICE:output)
        :return False if channel was not registered
        """
        return self.__client.unregisterChannelMonitor(channelName)

    def setDeviceMonitorInterval(self, milliseconds):
        """Set wait interval between subsequent handlers

        Set interval to wait between subsequent (for the same instance)
        calls to handlers registered via registerDeviceMonitor.
        Changes received within that interval will be cached and, in case of
        several updates of the same property within the interval, only the most
        up-to-date value will be handled.
        If negative, switch off caching and call handler immediately.
        """
        self.__client.setDeviceMonitorInterval(milliseconds)

    def registerPropertyMonitor(self, instanceId, propertyName,
                                callbackFunction, userData=None):
        """Register an asynchronous call-back on change of property

        This function can be used to register an asynchronous call-back on
        change of the specified property. The call-back function must have the
        following signature: f(str, str, object, Timestamp)
        arg1: deviceId
        arg2: key
        arg3: value
        arg4: timeStamp

        Example:

        def onPropertyChange(deviceId, key, value, timeStamp):
            print( deviceId, ":", key, "->", value, "(",
                   timeStamp.toFormattedString(), ")" )

        c = DeviceClient()
        c.registerPropertyMonitor("Test_Device_0", "result", onPropertyChange)
        """
        if userData is None:
            return self.__client.registerPropertyMonitor(
                instanceId, propertyName, callbackFunction)
        else:
            return self.__client.registerPropertyMonitor(
                instanceId, propertyName, callbackFunction, userData)

    def unregisterPropertyMonitor(self, instanceId, propertyName):
        self.__client.unregisterPropertyMonitor(instanceId, propertyName)

    def set(self, instanceId, propertyName, propertyValue, timeoutInSeconds=-1,
            keySep="."):
        return self.__client.set(instanceId, propertyName, propertyValue,
                                 keySep, timeoutInSeconds)

    def setHash(self, instanceId, hash, timeoutInSeconds=-1):
        return self.__client.set(instanceId, hash, timeoutInSeconds)

    def setNoWait(self, instanceId, propertyName, propertyValue):
        return self.__client.setNoWait(instanceId, propertyName, propertyValue)

    def execute(self, instanceId, command):
        """Executes a command"""
        return self.__client.execute(instanceId, command)

    def executeN(self, instanceId, command, *args):
        """Executes a command with args and return values"""
        return self.__client.executeN(instanceId, command, *args)

    def executeNoWait(self, deviceId, command):
        """Executes a command without waiting"""
        self.__client.executeNoWait(deviceId, command)

    def sleep(self, secs):
        time.sleep(secs)

    def _fromTimeStringToUtcString(self, timestamp):
        date = parser.parse(timestamp)
        if date.tzname() is None:
            print("Assuming local time for given date ", date)
            local_tz = tzlocal.get_localzone()
            date = local_tz.localize(date)
            date = date.astimezone(pytz.utc)
        print(date.isoformat())
        return date.isoformat()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import asyncio
import logging
import os
import re
from contextlib import asynccontextmanager, contextmanager
from functools import reduce, wraps
from time import perf_counter
from types import MethodType
from weakref import WeakMethod, ref

from karabo import __version__ as karabo_version
from karabo.native import Hash

_SCHEME_DELIMITER = "://"
_BROKER_REGEX = re.compile(
    r"^(amqp:\/\/)([A-Za-z0-9]+)(\:)+([A-Za-z0-9]+)"
    r"(\@)+([A-Za-z0-9\-\.]+)(\:)+([0-9]+)")


def ensure_coroutine(coro):
    """Ensure that a function `coro` is a coroutine to play well
    in our eventloops"""
    if asyncio.iscoroutinefunction(coro):
        return coro

    def create_coroutine(func):
        """Create a coroutine wrapper around the sync function `func`"""

        @wraps(func)
        async def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return create_coroutine(coro)


def get_karabo_version():
    """Return the karabo version
    """
    if "KARABO" not in os.environ:
        print("ERROR: $KARABO is not defined. Make sure you have sourced "
              "the 'activate' script.")
    return karabo_version


def get_property_hash(device, properties):
    """Get all properties from `device`

    :param device: The device instance or proxy object
    :param properties: The sequence of properties ('a', 'b') ...

    returns: Hash
    """
    ret = Hash()
    for key in properties:
        value = get_property(device, key)
        desc = value.descriptor
        ret.setElement(key, *desc.toDataAndAttrs(value))
    return ret


def get_property(device, path):
    """Return the property value from a proxy or device

    :param device: The device instance or proxy object
    :param path: The full path of the property as string

    This function is similar to python's builtin ``getattr`` and used with::

        prop = get_property(proxy, 'node.subnode.property')

    which is equivalent to::

        prop = proxy.node.subnode.property
    """
    return reduce(lambda obj, key: getattr(obj, key), path.split('.'), device)


def set_property(device, path, value):
    """Set a property value on a proxy or device

    This function has been added in Karabo >= 2.14.

    :param device: The device instance or proxy object
    :param path: The full path of the property as string
    :param value: The value to be set

    This function is similar to python's builtin ``setattr`` and used with::

        set_property(proxy, 'node.subnode.property', 5)

    which is equivalent to::

        proxy.node.subnode.property = 5
    """
    obj = device
    paths = path.split(".")
    for name in paths[:-1]:
        if not hasattr(obj, name):
            raise AttributeError(
                f"Property {path} is not available on device.")
        obj = getattr(obj, name)
    if not hasattr(obj, paths[-1]):
        raise AttributeError(f"Property {path} is not available on device.")
    setattr(obj, paths[-1], value)


class profiler:
    """A versatile profiling class

    Use this class either as a context manager or as decorator::

        with profiler():
            # classic context manager

        with profiler("Long computation 1"):
            # do something but provide a name for the context

        The class can also be used for decoration of functions (async works)

        @profiler()
        async def do_something()
            # do something

        async with profiler("Async with statement"):
            # do something with name in context

    """

    def __init__(self, name=None):
        self.name = name
        self.t_start = None

    def __enter__(self):
        self.t_start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed = perf_counter() - self.t_start
        name = f"{self.name}:" if self.name is not None else ":"
        print(f"With block {name} time elapsed {elapsed}")

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.__exit__(exc_type, exc_val, exc_tb)

    def __call__(self, func):
        """Decorate a function to profile the execution time"""
        name = func.__name__ if self.name is None else self.name
        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                t_start = perf_counter()
                ret = await func(*args, **kwargs)
                elapsed = perf_counter() - t_start
                print(f"{name} took {elapsed}")
                return ret
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                t_start = perf_counter()
                ret = func(*args, **kwargs)
                elapsed = perf_counter() - t_start
                print(f"{name} took {elapsed}")
                return ret

        return wrapper


class AsyncTimer:
    """AsyncTimer class to periodically execute callbacks

    this timer will be deleted once the instance creating it is being deleted.
    :param callback: Callable function to execute. If callback is a bound
                     method, the timer automatically creates a weak
                     reference
    :param timeout: Interval in seconds after which the callback is put on
                    the eventloop
    :param single_shot: Boolean to set if the timer only does a single call
    :param flush_interval: The "ultimate" interval in seconds after which the
                           `callback` will be called even if the timer is
                           restarted (i.e. "snoozed").
                           Default is `None` and implies that a timer can
                           be snoozed indefinitely.

    The timer starts and always defers if started again (snooze behavior)
    """

    def __init__(self, callback, timeout=0, single_shot=False,
                 flush_interval=None, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        instance = loop.instance()
        if instance is None:
            raise RuntimeError("A timer needs an instance to run ...")
        self.instance = ref(instance)
        self._register(instance)

        assert callable(callback)
        if type(callback) is MethodType:
            callback = WeakMethod(callback)

        self._callback = callback
        self._timeout = timeout
        self._single_shot = single_shot
        msg = f"Unexpected flush_interval: {flush_interval}"
        assert flush_interval is None or flush_interval > 0, msg
        self._flush_interval = flush_interval

        # Internally used variables
        self._handle = None
        self._time = None

    def _register(self, instance):
        """Internal method to register a timer on an `instance`"""
        instance._timers.add(self)
        weak_timer = ref(self)

        @contextmanager
        def destroyer():
            try:
                yield
            finally:
                weak = weak_timer()
                if weak is not None:
                    weak.destroy()

        instance._sigslot.enter_context(destroyer())

    def start(self):
        """Start the async timer

        If the timer was already started, the callback is postponed by a
        timeout interval.

        :returns bool: True if started, False otherwise
        """
        # We restarted ourselves, but now the loop is closed. We don't start
        # again.
        if self.loop.is_closed():
            return False
        if self._handle is None:
            self._time = self.loop.time()
            self._handle = self.loop.call_later(
                self._timeout,
                callback=self._channel_callback)
        else:
            if (self._flush_interval is not None and
                    self.loop.time() - self._time >=
                    self._flush_interval):
                # We don't allow postponing the callback anymore
                return
            self._handle.cancel()
            self._handle = self.loop.call_later(
                self._timeout,
                callback=self._channel_callback)
        return True

    def _channel_callback(self):
        """Private method to launch the callback as a task on the eventloop"""
        callback = (self._callback() if isinstance(self._callback, WeakMethod)
                    else self._callback)
        if callback is None:
            # destroyed or the object holding the WeakMethod is deleted
            self._handle = None
            self._time = None
            return

        coro = ensure_coroutine(callback)
        self.loop.call_soon_threadsafe(self.loop.create_task, coro(),
                                       self.instance())
        self._handle = None
        self._time = None
        if not self._single_shot:
            self.start()

    def is_running(self):
        """Check if the timer is running"""
        return self._handle is not None

    def stop(self):
        """Stop the execution of an asynchronous timer

        :returns bool: True if actively stopped, False otherwise
        """
        if self._handle is not None:
            self._time = None
            self._handle.cancel()
            self._handle = None
            return True
        return False

    def __del__(self):
        self.destroy()

    def destroy(self):
        """Stop the timer and delete any callback or loop reference"""
        self.stop()
        self._callback = None
        self.loop = None


@asynccontextmanager
async def countdown(delay=5, exception=True):
    """An async contextmanager to enter a timeout context

    :param delay: timeout limit (delay), default is 5 seconds
    :param exception: Boolean if an exception should be raised on timeout,
                      defaults to `True`
    """
    try:
        async with asyncio.timeout(delay) as ctx:
            yield ctx
    except TimeoutError:
        if exception:
            raise


class suppress:
    """A contextmanager to suppress exceptions in a context

    Note: On exception, a log message to the instance logger is generated
    """

    def __init__(self, *exceptions):
        self._exceptions = exceptions

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, exc_tb):
        suppressed = exc_type is not None and issubclass(
            exc_type, self._exceptions)
        if suppressed:
            try:
                logger = asyncio.get_event_loop().instance().logger
            except BaseException:
                # Make absolutely sure that this the log message is done!
                logger = logging.getLogger()
            text = "Suppressed exception ..."
            logger.error(text, exc_info=(exc_type, exc_value, exc_tb))

        return suppressed


def check_broker_scheme(urls: list):
    """Check the list of broker hosts

    :param hosts: The list of broker hosts

    This function will raise a `RuntimeError` if the provided list
    contains invalid broker host data.
    """

    def _all_equal(iterable):
        """Check if all elements of an iterable are equal"""
        iterator = iter(iterable)
        try:
            first = next(iterator)
        except StopIteration:
            return True
        return all(first == x for x in iterator)

    def _all_match(iterable, regex):
        """Check if all iterables match the regex"""
        iterator = iter(iterable)
        return all(regex.match(x) for x in iterator)

    if not len(urls):
        raise RuntimeError(
            "The list of broker hosts does contain any url.")

    schemes = [url.split(_SCHEME_DELIMITER)[0] for url in urls]
    if not len(schemes) or not all([scheme in "amqp" for scheme in schemes]):
        raise RuntimeError(
            "The list of broker hosts does provide unknown schemes.")

    if not _all_equal(schemes):
        raise RuntimeError(
            "The list of broker hosts contains different schemes.")

    if not _all_match(urls, _BROKER_REGEX):
        raise RuntimeError(
            "Incomplete broker schemes for amqp provided. It must be "
            "assembled like `amqp://user:password@host:port`.")



import pytest
from traits.api import Instance, List, String

from karabo.common.scenemodel.api import (
    BaseROIData, BaseWidgetObjectData, CrossROIData, RectROIData, SceneModel)
from karabo.interactive import scene2python


class SimpleModel(BaseWidgetObjectData):
    name = String
    rois = List(Instance(BaseROIData))


class MultiChildren(BaseWidgetObjectData):
    name = String
    cross_rois = List(Instance(CrossROIData))
    rect_rois = List(Instance(RectROIData))


@pytest.fixture
def models():
    cross_roi1 = CrossROIData(name="FirstCrossROI", x=10, y=10)
    cross_roi2 = CrossROIData(name="SecondCrossROI", x=2, y=7)

    rect_roi1 = RectROIData(name="FirstRectROI", x=6, y=21)
    rect_roi2 = RectROIData(name="SecondRectROI", x=12, y=43)

    rois = [cross_roi1, cross_roi2, rect_roi1, rect_roi2]
    cross_rois = [cross_roi1, cross_roi2]
    rect_rois = [rect_roi1, rect_roi2]

    # Single Children
    simple = SimpleModel(name="only_one_child", rois=rois)

    # Mulitple children (More than one list trait)
    multi_child = MultiChildren(name="multiple_children",
                                cross_rois=cross_rois,
                                rect_rois=rect_rois)
    return simple, multi_child


def test_code_from_simple_model(models):
    simple, multi_child = models
    classes, code = scene2python.convert_scene_model_to_code(simple)
    assert classes == {"CrossROIData", "RectROIData", "SimpleModel"}
    expected_code = [
        "scene00 = CrossROIData(name='FirstCrossROI', x=10.0, y=10.0)",
        "scene01 = CrossROIData(name='SecondCrossROI', x=2.0, y=7.0)",
        "scene02 = RectROIData(name='FirstRectROI', x=6.0, y=21.0)",
        "scene03 = RectROIData(name='SecondRectROI', x=12.0, y=43.0)",
        ("scene = SimpleModel(name='only_one_child', "
         "parent_component='DisplayComponent', rois=[scene00, scene01, "
         "scene02, scene03])")
    ]
    assert code == expected_code

    # With Multiple Children
    classes, complex_code = scene2python.convert_scene_model_to_code(
        multi_child)
    assert classes == {"CrossROIData", "RectROIData", "MultiChildren"}
    expected_code = [
        "scene00 = CrossROIData(name='FirstCrossROI', x=10.0, y=10.0)",
        "scene01 = CrossROIData(name='SecondCrossROI', x=2.0, y=7.0)",
        "scene10 = RectROIData(name='FirstRectROI', x=6.0, y=21.0)",
        "scene11 = RectROIData(name='SecondRectROI', x=12.0, y=43.0)",
        ("scene = MultiChildren(name='multiple_children', "
         "parent_component='DisplayComponent', cross_rois=[scene00, "
         "scene01], rect_rois=[scene10, scene11])")
    ]

    assert expected_code == complex_code


def test_scene_model(models):
    # Within a Scene model.
    simple, multi_child = models
    simple_scene_model = SceneModel(children=[simple])
    classes, code = scene2python.convert_scene_model_to_code(
        simple_scene_model)
    assert classes == {'CrossROIData', 'RectROIData', 'SceneModel',
                       'SimpleModel'}
    expected = [
        "scene0000 = CrossROIData(name='FirstCrossROI', x=10.0, y=10.0)",
        "scene0001 = CrossROIData(name='SecondCrossROI', x=2.0, y=7.0)",
        "scene0002 = RectROIData(name='FirstRectROI', x=6.0, y=21.0)",
        "scene0003 = RectROIData(name='SecondRectROI', x=12.0, y=43.0)",
        ("scene00 = SimpleModel(name='only_one_child', "
         "parent_component='DisplayComponent', rois=[scene0000, scene0001, "
         "scene0002, scene0003])"),
        'scene = SceneModel(children=[scene00])']
    assert code == expected



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from contextlib import contextmanager

from karabo.middlelayer import (
    KaraboError, get_property, getDevice, setWait, updateDevice)
from karabo.middlelayer.tests.eventloop import DeviceTest, async_tst

from ..property_test import PropertyTest

conf = {
    "classId": "PropertyTest",
    "_deviceId_": "Test_PropertyTest"
}


class Tests(DeviceTest):
    @classmethod
    @contextmanager
    def lifetimeManager(cls):
        cls.dev = PropertyTest(conf)
        with cls.deviceManager(lead=cls.dev):
            yield

    @async_tst
    async def test_property_test(self):
        # tests are run in sequence as sub tests
        # device server thus is only instantiated once

        with self.subTest(msg="Test setting read-only"):
            props = ["boolProperty",
                     "int32Property", "uint32Property",
                     "int64Property", "uint64Property",
                     "floatProperty", "doubleProperty",
                     "node.counter"]

            with (await getDevice("Test_PropertyTest")) as d:
                await updateDevice(d)
                for i, prop in enumerate(props, 10):
                    self.assertNotEqual(get_property(d, prop), i)
                    await setWait(d, prop, i)
                    self.assertEqual(get_property(d, prop), i)
                    self.assertEqual(get_property(d, prop + "ReadOnly"), i)

        with self.subTest(msg="Test alarm slots"):
            with (await getDevice("Test_PropertyTest")) as d:
                await updateDevice(d)
                # invalid alarm condition in stringProperty
                await setWait(d, "stringProperty", "")
                with self.assertRaises(KaraboError):
                    await d.setAlarm()

                # valid alarm condition in stringProperty
                await setWait(d, "stringProperty", "warn")
                await d.setAlarm()
                self.assertEqual(d.alarmCondition.value, "warn")

        with self.subTest(msg="Test noded slots"):
            with (await getDevice("Test_PropertyTest")) as d:
                await updateDevice(d)
                await d.node.reset()
                self.assertEqual(d.node.counterReadOnly, 0)
                await d.node.increment()
                self.assertEqual(d.node.counterReadOnly, 1)
                await d.node.reset()
                self.assertEqual(d.node.counterReadOnly, 0)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from enum import Enum
from textwrap import dedent
from unittest import TestCase, main, skip
from xml.etree.ElementTree import fromstring, tostring

from IPython.lib.pretty import pretty

from karabo.middlelayer import (
    Float, Hash, Int32, Proxy, ProxyNode, SubProxy, Unit, VectorString)


class Node(SubProxy):
    strings = VectorString(key="strings")

    _allattrs = ["strings"]


class Words(Enum):
    a = 5
    b = 7


class Sample(Proxy):
    bin_desc = Int32(
        displayType="""bin|1:a,2:some really long description
                    for testing if line < break works properly,3:c""",
        key="bin_desc")
    bin_int = Int32(displayType="bin", key="bin_int")
    oct_int = Int32(displayType="oct", key="oct_int")
    hex_int = Int32(displayType="hex", key="hex_int")
    enum_int = Int32(enum=Words, key="enum_int")
    ohm_float = Float(unitSymbol=Unit.OHM, key="ohm_float")
    err_float = Float(absoluteError=0.1, relativeError=0.0001, key="err_float")
    strings = VectorString(key="strings")
    node = ProxyNode(cls=Node, key="node")

    _allattrs = ["bin_desc", "bin_int", "oct_int", "hex_int", "enum_int",
                 "ohm_float", "err_float", "strings", "node"]


class Tests(TestCase):
    def assertPretty(self, challenge, response):
        self.assertEqual(pretty(challenge), dedent(response[1:]))

    def assertHtml(self, challenge, response):
        challenge = tostring(fromstring(challenge._repr_html_()),
                             encoding="unicode")
        response = "".join(line.strip().replace("$", "\n")
                           for line in response.split("\n"))
        self.assertEqual(challenge, response)

    @skip
    def test_none(self):
        s = Sample()
        self.assertPretty(s, """

            bin_desc: None
            bin_int: None
            oct_int: None
            hex_int: None
            enum_int: None
            ohm_float: None
            err_float: None
            strings: None
            node:
                strings: None""")

        self.assertHtml(s, """
            <table>
                <tr>
                    <td style="padding-left:1em">bin_desc</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em">bin_int</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em">oct_int</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em">hex_int</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em">enum_int</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em">ohm_float</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em">err_float</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em">strings</td>
                    <td>None</td>
                </tr><tr>
                    <td style="padding-left:1em"><b>node</b></td>
                    <td />
                </tr><tr>
                    <td style="padding-left:2em">strings</td>
                    <td>None</td>
                </tr>
            </table>
        """)

    @skip
    def test_empty(self):
        s = Sample()
        s._onChanged(Hash("bin_desc", 0, "bin_int", 0, "oct_int", 0,
                          "hex_int", 0, "enum_int", 5, "strings", [],
                          "ohm_float", 0, "err_float", 0,
                          "node.strings", []))
        self.assertPretty(s, """

            bin_desc: {  }
            bin_int: 0b0
            oct_int: 0o0
            hex_int: 0x0
            enum_int: <Words.a>
            ohm_float: 0.0 Ω
            err_float: 0.0
            strings: []
            node:
                strings: []""")

        self.assertHtml(s, """
            <table>
                <tr>
                    <td style="padding-left:1em">bin_desc</td>
                    <td />
                </tr><tr>
                    <td style="padding-left:1em">bin_int</td>
                    <td>0b0</td>
                </tr><tr>
                    <td style="padding-left:1em">oct_int</td>
                    <td>0o0</td>
                </tr><tr>
                    <td style="padding-left:1em">hex_int</td>
                    <td>0x0</td>
                </tr><tr>
                    <td style="padding-left:1em">enum_int</td>
                    <td><i>Words.a</i></td>
                </tr><tr>
                    <td style="padding-left:1em">ohm_float</td>
                    <td>0.0 Ω</td>
                </tr><tr>
                    <td style="padding-left:1em">err_float</td>
                    <td>0.0 </td>
                </tr><tr>
                    <td style="padding-left:1em">strings</td>
                    <td />
                </tr><tr>
                    <td style="padding-left:1em"><b>node</b></td>
                    <td />
                </tr><tr>
                    <td style="padding-left:2em">strings</td>
                    <td />
                </tr>
            </table>
        """)

    @skip
    def test_short(self):
        s = Sample()
        s._onChanged(Hash("bin_desc", 2, "bin_int", 5, "oct_int", 10,
                          "hex_int", 20, "enum_int", 5, "strings", ["a"],
                          "ohm_float", 2.000001, "err_float", 3.1415926,
                          "node.strings", ["b"]))
        self.assertPretty(s, """

            bin_desc: { a }
            bin_int: 0b101
            oct_int: 0o12
            hex_int: 0x14
            enum_int: <Words.a>
            ohm_float: 2.000001 Ω
            err_float: 3.1
            strings: [a]
            node:
                strings: [b]""")

        self.assertHtml(s, """
            <table>
                <tr>
                    <td style="padding-left:1em">bin_desc</td>
                    <td>a</td>
                </tr><tr>
                    <td style="padding-left:1em">bin_int</td>
                    <td>0b101</td>
                </tr><tr>
                    <td style="padding-left:1em">oct_int</td>
                    <td>0o12</td>
                </tr><tr>
                    <td style="padding-left:1em">hex_int</td>
                    <td>0x14</td>
                </tr><tr>
                    <td style="padding-left:1em">enum_int</td>
                    <td><i>Words.a</i></td>
                </tr><tr>
                    <td style="padding-left:1em">ohm_float</td>
                    <td>2.000001 Ω</td>
                </tr><tr>
                    <td style="padding-left:1em">err_float</td>
                    <td>3.1 </td>
                </tr><tr>
                    <td style="padding-left:1em">strings</td>
                    <td>a</td>
                </tr><tr>
                    <td style="padding-left:1em"><b>node</b></td>
                    <td />
                </tr><tr>
                    <td style="padding-left:2em">strings</td>
                    <td>b</td>
                </tr>
            </table>
        """)

    @skip
    def test_long(self):
        s = Sample()
        s._onChanged(Hash("bin_desc", 6, "bin_int", -1000000000,
                          "oct_int", -2000000000,
                          "hex_int", -1500000000,
                          "enum_int", 5,
                          "ohm_float", 1.000001e-7,
                          "err_float", 12345678,
                          "strings", [
                              "this", """is a list of long strings which will
                            most likely exceed one line"""],
                          "node.strings", ["another", "list", "this", "time",
                                           "with", "many", "ent>ries", "also",
                                           "exceeding", "one", "line"]))
        self.assertPretty(s, """

            bin_desc:
                { a
                | some really long description
                                for testing if line < break works properly }
            bin_int: 0b-111011100110101100101000000000
            oct_int: 0o-16715312000
            hex_int: 0x-59682f00
            enum_int: <Words.a>
            ohm_float: 1.000001e-07 Ω
            err_float: 1.2346e+07
            strings:
                [this,
                 is a list of long strings which will
                                        most likely exceed one line]
            node:
                strings:
                    [another,
                     list,
                     this,
                     time,
                     with,
                     many,
                     ent>ries,
                     also,
                     exceeding,
                     one,
                     line]""")

        self.assertHtml(s, """
            <table>
                <tr>
                    <td style="padding-left:1em">bin_desc</td>
                    <td>a<br />some really long des
                        cription$                    for test
                        ing if line &lt; break works properly</td>
                </tr><tr>
                    <td style="padding-left:1em">bin_int</td>
                    <td>0b-111011100110101100101000000000</td>
                </tr><tr>
                    <td style="padding-left:1em">oct_int</td>
                    <td>0o-16715312000</td>
                </tr><tr>
                    <td style="padding-left:1em">hex_int</td>
                    <td>0x-59682f00</td>
                </tr><tr>
                    <td style="padding-left:1em">enum_int</td>
                    <td><i>Words.a</i></td>
                </tr><tr>
                    <td style="padding-left:1em">ohm_float</td>
                    <td>1.000001 × 10<sup>-7</sup> Ω</td>
                </tr><tr>
                    <td style="padding-left:1em">err_float</td>
                    <td>1.2346 × 10<sup>7</sup> </td>
                </tr><tr>
                    <td style="padding-left:1em">strings</td>
                    <td>this<br />
                        is a list of long strings which wi
                        ll$                            most likely ex
                        ceed one line</td>
                </tr><tr>
                    <td style="padding-left:1em"><b>node</b></td>
                    <td />
                </tr><tr>
                    <td style="padding-left:2em">strings</td>
                    <td>another<br />list<br />this<br />time<br />
                        with<br />many<br />ent&gt;ries<br />also<br />
                        exceeding<br />one<br />line</td>
                </tr>
            </table>
        """)


if __name__ == "__main__":
    main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import getpass
import json
import uuid
from asyncio import gather, sleep, wait_for

import pytest

from karabo.middlelayer.device import Device
from karabo.middlelayer.device_client import (
    call, getClassSchema, getInstanceInfo, instantiateNoWait, waitUntil)
from karabo.middlelayer.testing import (
    create_device_server, create_instanceId, sleepUntil)
from karabo.native import Hash, KaraboError, Schema, Timestamp

SHUTDOWN_TIME = 1

FLAKY_MAX_RUNS = 3
FLAKY_MIN_PASSES = 1


class FaultyDevice(Device):

    async def onInitialization(self):
        raise RuntimeError("Not allowed to start")


@pytest.mark.flaky(max_runs=FLAKY_MAX_RUNS, min_passes=FLAKY_MIN_PASSES)
@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_device_server_no_plugins():
    serverId = f"testMDLServer-{uuid.uuid4()}"
    configuration = {"timeServerId": "KaraboTimeServer",
                     "track": True}
    server = create_device_server(serverId, config=configuration)
    assert server is not None
    try:
        server.is_server = True
        await server.startInstance(broadcast=True)

        info = await getInstanceInfo(serverId)
        # Check the serverFlags, None set
        assert info["serverFlags"] == 0

        # And the time info
        th = server.slotGetTime(Hash())
        keys = list(th.keys())
        assert "reference" in keys
        ts_ref = Timestamp.fromHashAttributes(th["reference", ...])
        assert ts_ref.tid == 0

        assert "time" in keys
        assert "timeServerId" in keys
        assert th["timeServerId"] == "KaraboTimeServer"

        t = Timestamp()
        server.slotTimeTick(100, t.time_sec, t.time_frac, 200)
        th = server.slotGetTime(Hash())
        ts_ref = Timestamp.fromHashAttributes(th["reference", ...])
        assert ts_ref.tid == 100
        assert ts_ref.time_sec == t.time_sec
        assert ts_ref.time_frac == t.time_frac

        # Remove start ticks again
        server.slotTimeTick(0, t.time_sec, t.time_frac, 200)
    finally:
        await finalize_server(server)


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_device_server_instantiate_plugins():
    serverId = f"testMDLServer-{uuid.uuid4()}"

    configuration = {"deviceClasses": ["PropertyTest"],
                     "timeServerId": "KaraboTimeServer",
                     "serverFlags": ["Development"]}
    server = create_device_server(serverId, config=configuration)
    assert server is not None
    try:
        server.is_server = True
        await server.startInstance(broadcast=True)
        assert server.deviceClasses == ["PropertyTest"]

        assert len(server.deviceInstanceMap.keys()) == 0
        # Test instantiate a device
        deviceId = f"test-mdlfake-{uuid.uuid4()}"
        await server.startDevice("PropertyTest", deviceId,
                                 Hash())
        assert len(server.deviceInstanceMap.keys()) == 1
        classes = server.getClasses()
        assert classes == ["PropertyTest"]

        schema, classId, serv_id = server.slotGetClassSchema(
            "PropertyTest")
        assert serv_id == serverId
        assert classId == "PropertyTest"
        assert isinstance(schema, Schema)

        info = await getInstanceInfo(serverId)
        # Check the serverFlags
        assert info["serverFlags"] == 1

        # And the log level
        assert info["log"] == "INFO"
        assert server.log.level == "INFO"
        device = list(server.deviceInstanceMap.values())[0]
        assert device.log.level == "INFO"
        # Change log level
        server.slotLoggerLevel("ERROR")
        assert server.log.level == "ERROR"
        assert device.log.level == "ERROR"
        # Provide user
        assert info["user"] == getpass.getuser()

        info = await getInstanceInfo(serverId)
        assert info["log"] == "ERROR"

        logs = await call(serverId, "slotLoggerContent", Hash())
        assert logs["serverId"] == serverId
        assert isinstance(logs["content"], list)
        logs = await call(serverId, "slotLoggerContent",
                          Hash("logs", 20))
        assert isinstance(logs["content"], list)

        schema = await getClassSchema(serverId, "PropertyTest")
        assert schema is not None
        assert isinstance(schema, Schema)
    finally:
        await finalize_server(server)


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_device_server_autostart():
    deviceId_1 = f"test-prop-{uuid.uuid4()}"
    deviceId_2 = f"test-prop-{uuid.uuid4()}"
    serverId = f"testMDLServer-{uuid.uuid4()}"

    init = {deviceId_1: {"classId": "PropertyTest"},
            deviceId_2: {"classId": "PropertyTest"}}
    init = json.dumps(init)

    configuration = {"deviceClasses": ["PropertyTest"],
                     "timeServerId": "KaraboTimeServer"}
    server = create_device_server(serverId, config=configuration)
    assert server is not None
    try:
        server.is_server = True
        server._device_initializer = json.loads(init)
        await server.startInstance(broadcast=True)
        # Wait until a device has been registered in this server
        await waitUntil(lambda: len(server.deviceInstanceMap) == 2)
        assert server.deviceClasses == ["PropertyTest"]
        assert len(server.deviceInstanceMap.keys()) == 2
        assert deviceId_1 in server.deviceInstanceMap
        assert deviceId_2 in server.deviceInstanceMap
        classes = server.getClasses()
        assert classes == ["PropertyTest"]
        device = server.deviceInstanceMap[deviceId_1]
        assert server == device.device_server
        local = device.getLocalDevice(deviceId_2)
        assert local is not None
        assert local.classId == "PropertyTest"
        local = device.getLocalDevice("Notthere")
        assert local is None
    finally:
        await finalize_server(server)


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_device_server_start_faulty():
    try:
        serverId = create_instanceId()
        server = create_device_server(serverId,
                                      plugins=[FaultyDevice])
        await server.startInstance()
        await waitUntil(lambda: server.is_initialized)
        deviceId = create_instanceId()
        hsh = Hash(
            "classId", "FaultyDevice",
            "deviceId", deviceId,
            "configuration", Hash())

        # Succesful instantiation although broken device
        ok, *_ = await server.slotStartDevice(hsh)
        assert ok is True
        # Device is not there in server map
        assert deviceId not in server.deviceInstanceMap

        # This is concurrent instantiation, which fails because
        # of exclusive queue use in broker setup.
        with pytest.raises(KaraboError):
            fut = [call(server.serverId,
                        "slotStartDevice", hsh) for _ in range(2)]
            await gather(*fut)

        # This is another concurrent instantiation to test the ramp down
        # with ordered subscriptions (broker)
        hsh = Hash(
            "classId", "FaultyDevice",
            "deviceId", "AnotherdeviceId",
            "configuration", Hash())
        with pytest.raises(RuntimeError):
            fut = [server.slotStartDevice(hsh) for _ in range(2)]
            await gather(*fut)
    finally:
        await finalize_server(server)


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_device_server_concurrence():
    try:
        serverId = create_instanceId()
        configuration = {"deviceClasses": ["PropertyTest"]}
        server = create_device_server(serverId, config=configuration)
        await server.startInstance()
        await waitUntil(lambda: server.is_initialized)
        deviceId = create_instanceId()
        assert len(server.deviceInstanceMap) == 0
        for _ in range(2):
            instantiateNoWait(serverId, "PropertyTest", deviceId)
        # Only a single device made it
        await sleepUntil(lambda: len(server.deviceInstanceMap) > 0,
                         timeout=10)
        assert len(server.deviceInstanceMap) == 1
    finally:
        await finalize_server(server)


async def finalize_server(server):
    """Helper function to stop the device server

    Note: Stop the server with slotKillDevice, but not the eventloop
    Make sure all children devices are shutdown!
    """
    futures = [d.slotKillDevice()
               for d in server.deviceInstanceMap.values()]
    if futures:
        await wait_for(gather(*futures), timeout=10)
    await server.slotKillDevice()
    await sleep(SHUTDOWN_TIME)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from ..const import (
    ROOT_COLLECTION, ROOT_COLLECTION_BACKUP, ROOT_COLLECTION_TEST)


class DbSettings:
    def __init__(self, user, password, server, port=8080, init_db=False):
        self.user = user
        self.password = password
        self.server = server
        self.port = port
        self.init_db = init_db
        self.server_uri = make_server_uri(server, port)
        self.server_url = make_server_url(self.user,
                                          self.password,
                                          self.server_uri)
        self.root_collection = f"/{ROOT_COLLECTION}"
        self.root_collection_test = f"/{ROOT_COLLECTION_TEST}"
        self.root_collection_backup = f"/{ROOT_COLLECTION_BACKUP}"


def make_server_uri(server, port):
    return f"{server}:{port}/exist/xmlrpc"


def make_server_url(user, password, server_uri):
    return f"http://{user}:{password}@{server_uri}"



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from asyncio import get_running_loop

import pytest
import pytest_asyncio

from karabo.middlelayer import (
    DeviceClientBase, VectorString, listCameras, listDeviceInstantiators,
    listMotors, listProcessors, listTriggers)
from karabo.middlelayer.testing import AsyncDeviceContext, run_test


class MyDevice(DeviceClientBase):
    interfaces = VectorString(
        defaultValue=["Motor", "Camera", "Processor", "DeviceInstantiator"])


@pytest_asyncio.fixture(scope="module")
async def deviceTest():
    myDevice = MyDevice(dict(_deviceId_="MyDevice"))
    get_running_loop().lead = myDevice
    async with AsyncDeviceContext(myDevice=myDevice) as ctx:
        yield ctx


@pytest.mark.timeout(30)
@run_test
async def test_find_interfaces(deviceTest):
    must_have = ["Motor", "Camera", "Processor", "DeviceInstantiator"]
    assert deviceTest["myDevice"].interfaces == must_have
    motors = listMotors()
    assert "MyDevice" in motors
    cameras = listCameras()
    assert "MyDevice" in cameras
    cameraName = listCameras("MYDEVICE")
    assert "MyDevice" in cameraName
    cameraName = listCameras("mydevice")
    assert "MyDevice" in cameraName
    triggers = listTriggers()
    assert "MyDevice" not in triggers
    processors = listProcessors()
    assert "MyDevice" in processors
    # Find rubbish
    motors = listMotors("deep")
    assert "MyDevice" not in motors
    instantiator = listDeviceInstantiators("mydevice")
    assert "MyDevice" in instantiator



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import asyncio
import time
from collections import defaultdict
from contextlib import suppress
from inspect import isfunction
from weakref import WeakSet

from aiormq.exceptions import AMQPException

from karabo.common.api import WeakMethodRef
from karabo.middlelayer.synchronization import synchronize, synchronous
from karabo.native import (
    Descriptor, Hash, KaraboError, KaraboValue, NDArray, NodeType, NoneValue,
    Slot, Timestamp, Type, Weak, get_timestamp, isSet)

UNSUBSCRIBE_TIMEOUT = 5


class _ProxyBase:
    _parent = Weak()

    @classmethod
    def __dir__(cls):
        return cls._allattrs

    def __repr__(self):
        subs = ", ".join(f"{k}={getattr(self, k)!r}"
                         for k, v in self.__class__.__dict__.items()
                         if isinstance(v, Descriptor) and
                         not isinstance(v, Slot) and hasattr(self, k))
        return f"{self.__class__.__name__}[{subs}]"

    def _repr_html_generator_(self, nest=0):
        if nest == 0:
            yield "<table>"
        for attr in self._allattrs:
            val = getattr(self, attr, None)
            if isinstance(val, _ProxyBase):
                yield ('<tr><td style="padding-left:{}em"><b>{}</b></td>'
                       '<td/></tr>'.format(nest + 1, attr))
                yield from val._repr_html_generator_(nest + 1)
            elif isinstance(val, KaraboValue):
                yield ('<tr><td style="padding-left:{}em">{}</td><td>'
                       .format(nest + 1, attr))
                yield from val._repr_html_generator_()
                yield '</td></tr>'
        if nest == 0:
            yield "</table>"

    def _repr_html_(self):
        return "".join(self._repr_html_generator_())

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(str(self))
        else:
            for attr in self._allattrs:
                val = getattr(self, attr, None)
                if isinstance(val, (_ProxyBase, KaraboValue)):
                    p.break_()
                    with p.group(4, f"{attr}:"):
                        p.breakable()
                        p.pretty(val)

    def _getValue(self, key):
        self._parent._use()
        ret = self.__dict__.get(key)
        if ret is None:
            ret = NoneValue(descriptor=getattr(self.__class__, key))
            ret._parent = self._parent
        return ret


class ProxyBase(_ProxyBase):
    """The base for proxies to a Karabo device

    The :class:`ProxyFactory` will subclass this class and add
    the necessary descriptors.
    """

    def __init__(self):
        super().__init__()
        self._parent = self

    def _use(self):
        pass

    def _callSlot(self, descriptor):
        """this gets called once the user calls a slot"""

    def _onChanged_r(self, change, instance):
        for k, v, a in change.iterall():
            descr = getattr(type(instance), k, None)
            if descr is not None:
                if isinstance(descr, ProxyNodeBase):
                    self._onChanged_r(v, getattr(instance, descr.key))
                elif not isinstance(descr, ProxySlotBase):
                    converted = descr.toKaraboValue(v, strict=False)
                    converted.timestamp = Timestamp.fromHashAttributes(a)
                    converted._parent = self
                    instance.__dict__[descr.key] = converted
                    self._notifyChanged(descr, converted)

    def _onChanged_timestamp_r(self, change, timestamp, instance):
        """Recursively set data on a proxy with a fixed timestamp

        This function can be optionally used for pipeline data
        """
        for k, v, a in change.iterall():
            descr = getattr(type(instance), k, None)
            if descr is not None:
                if isinstance(descr, ProxyNodeBase):
                    self._onChanged_timestamp_r(
                        v, timestamp, getattr(instance, descr.key))
                elif not isinstance(descr, ProxySlotBase):
                    converted = descr.toKaraboValue(v, strict=False)
                    converted.timestamp = timestamp
                    converted._parent = self
                    instance.__dict__[descr.key] = converted
                    self._notifyChanged(descr, converted)

    def _onChanged(self, change):
        """call this when the remote device changed

        :change: is the Hash with the changes
        """
        self._onChanged_r(change, self)

    def configurationAsHash(self):
        """Extract a configuration in a Hash from the proxy

        Some values of a `ListOfNodes` appear in the configuration, but are not
        available on the proxy (for now).
        """
        ret = Hash()

        def recurse(proxy):
            nonlocal ret
            for key in proxy._allattrs:
                descr = getattr(type(proxy), key, None)
                if descr is None:
                    # Unsupported types
                    continue
                if isinstance(descr, ProxySlotBase):
                    # Slots don't appear in Configuration
                    continue
                elif isinstance(descr, ProxyNodeBase):
                    # recurse Nodes
                    recurse(getattr(proxy, descr.key))
                else:
                    value = getattr(proxy, key, None)
                    if isSet(value):
                        data, attrs = descr.toDataAndAttrs(value)
                        ret[descr.longkey] = data
                        ret[descr.longkey, ...].update(attrs)
        recurse(self)

        return ret

    def _notifyChanged(self, descriptor, value):
        """this is called by _onChanged for each change"""


class ProxyNodeBase(Descriptor):
    """The base class for all nodes in a Proxy

    This is the :class:`Descriptor` for a node in the class. The actual
    data's base class is :class:`SubProxyBase`.
    """

    def __init__(self, *, cls, **kwargs):
        super().__init__(**kwargs)
        self.cls = cls

    def __get__(self, instance, owner):
        if instance is None:
            return self
        ret = instance.__dict__.get(self.key, None)
        if ret is not None:
            return ret
        sub = self.cls()
        if isinstance(instance, SubProxyBase):
            sub._parent = instance._parent
        else:
            sub._parent = instance
        instance.__dict__[self.key] = sub
        return sub


class ProxySlotBase(Slot, Descriptor):
    """The base class for Slots in proxies"""

    def __get__(self, instance, owner):
        if instance is None:
            return self

        def method(myself, *args, **kwargs):
            return myself._parent._callSlot(self, *args, **kwargs)
        method.__doc__ = self.description
        return method.__get__(instance, owner)


class SubProxyBase(_ProxyBase):
    """The base class for nodes in a Proxy"""

    def setValue(self, desc, value):
        return self._parent.setValue(desc, value)


class ProxyFactory:
    """Create a proxy for Karabo devices

    This is an customizable factory class to create proxies for Karabo
    devices. You can customize which classes are used in the proxy
    building process by setting the following class attributes:

    Proxy
        This is the base class for all proxies. It should inherit from
        :class:`ProxyBase`.

    ProxySlot
        This is the base class for all slots of a device. It should inherit
        from :class:`ProxySlotBase`.

    ProxyNode
        The base class for all descriptors of nodes. Should inherit
        from :class:`ProxyNodeBase`.

    SubProxy
        This is the base class for all the node values in a proxy.
        It should inherit from :class:`SubProxyBase`.
    """
    Proxy = ProxyBase
    ProxyNode = ProxyNodeBase
    SubProxy = SubProxyBase

    node_factories = dict(
        Slot=ProxySlotBase,
        NDArray=NDArray)

    @classmethod
    def createNode(cls, key, node, prefix, **kwargs):
        sub = cls.createNamespace(node, f"{prefix}{key}.")
        Cls = type(key, (cls.SubProxy,), sub)
        return cls.ProxyNode(key=key, cls=Cls, **kwargs)

    @classmethod
    def register_special(cls, name, special):
        """register a special class *special* with *name*

        This adds a new handler for a classId. Note that this is registered
        in the class of the caller, and normal Python inheritance rules apply,
        so you can overload the registration in a specialized class.
        """
        cls.node_factories[name] = special

    @classmethod
    def createNamespace(cls, schema, prefix=""):
        namespace = {}
        for k, v, a in schema.iterall():
            nodeType = NodeType(a["nodeType"])
            if nodeType is NodeType.Leaf:
                descriptor = Type.fromname[a["valueType"]](
                    strict=False, key=k, **a)
                descriptor.longkey = prefix + k
            elif nodeType is NodeType.Node:
                classId = a.get("classId")
                factory = cls.node_factories.get(classId, cls.createNode)
                descriptor = factory(key=k, node=v, prefix=prefix,
                                     strict=False, **a)
            else:
                continue  # currently unsupported NodeType
            descriptor.longkey = prefix + k
            namespace[k] = descriptor
        namespace["_allattrs"] = list(schema)
        return namespace

    @classmethod
    def createProxy(cls, schema):
        namespace = cls.createNamespace(schema.hash)
        namespace["_schema_hash"] = schema.hash
        return type(schema.name, (cls.Proxy,), namespace)

    @classmethod
    def updateSchema(cls, proxy, schema):
        """update the schema of proxy"""
        def recurse(newcls, instance):
            instance.__class__ = newcls
            for key in dir(newcls):
                descriptor = getattr(newcls, key)
                if isinstance(descriptor, ProxyNodeBase):
                    value = getattr(instance, key, None)
                    if isinstance(value, SubProxyBase):
                        recurse(descriptor.cls, value)
                    else:
                        # something became a Node which wasn't one before.
                        # simply delete it to start anew at the next change
                        instance.__dict__.pop(key, None)
        newcls = cls.createProxy(schema)
        recurse(newcls, proxy)


class DeviceClientProxyFactory(ProxyFactory):
    class Proxy(ProxyBase):
        """A proxy for a remote device

        This proxy represents a real device somewhere in the karabo
        system. It is typically created by :func:`getDevice`, do not
        create it yourself.

        Properties and slots may be accessed as if one would access
        the actual device object, as in::

            device = getDevice("someDevice")
            device.speed = 7
            device.start()

        Note that in order to reduce network traffic, the proxy device
        is initially not *connected*, meaning that it does not receive
        updates of properties from the device. There are two ways to
        change that: you can call :func:`updateDevice` to get the
        current configuration once, or you may use the proxy in a with
        statment, which will connect the device during the execution
        of that statement, as in::

            with getDevice("someDevice") as device:
                print(device.speed)
        """

        def __init__(self, device, deviceId, sync):
            super().__init__()
            # device holding the proxy, e.g. DeviceClient (CLI) or
            # a middlelayer device
            self._device = device
            # None queue is global
            self._queues = defaultdict(WeakSet)
            # the deviceId
            self._deviceId = deviceId
            # variable used for autodisconnect
            self._used = 0
            # Hash container used for bulksetting
            self._sethash = Hash()
            # indicates if we are in sync with the eventloop
            self._sync = sync
            # our indicator if the underlying device is alive
            self._alive = True
            self._running_tasks = set()
            self._last_update_task = None
            self._schemaUpdateConnected = False
            self._lock_count = 0
            self._remote_output_channel = WeakSet()
            self._configurationHandler = None

        def _notifyChanged(self, descriptor, value):
            for q in self._queues[descriptor.longkey]:
                q.put_nowait(value)

        def _onChanged(self, change):
            if self._configurationHandler is not None:
                self._configurationHandler(change)
            super()._onChanged(change)
            for q in self._queues[None]:
                q.put_nowait(change)

        def setConfigHandler(self, changeHandler):
            """subscribe a configuration Change Handler

            This feature is supposed to be used for test and is experimental.
            Any feature built on it could break.
            Pass a synchronous function to it. To unsubscribe from it,
            pass a ``None``

            :changeHandler: a function taking a hash in input or a None
            """
            assert changeHandler is None or isfunction(changeHandler)
            self._configurationHandler = changeHandler

        def setValue(self, desc, value):
            """Set a value belonging to a descriptor on a proxy
            """
            self._use()
            loop = asyncio.get_event_loop()
            assert isinstance(value, KaraboValue)
            if loop.sync_set:
                # we are in a thread not running on Eventloop and
                # immediately send out our changes
                h = Hash()
                h[desc.longkey], _ = desc.toDataAndAttrs(value)
                loop.sync(self._raise_on_death(
                    self._device.call, self._deviceId, "slotReconfigure", h),
                    timeout=-1, wait=True)
            elif (self._last_update_task is not None and
                  self._last_update_task.done() and
                  self._last_update_task.exception() is not None):
                task = self._last_update_task
                self._last_update_task = None
                task.result()  # raise the exception
            else:
                # we are on the Eventloop and can do bulksetting
                update = not self._sethash
                self._sethash[desc.longkey], _ = desc.toDataAndAttrs(value)
                if update:
                    self._last_update_task = asyncio.ensure_future(
                        self._update(self._last_update_task))

        def _callSlot(self, descriptor, timeout=-1, wait=True):
            async def inner():
                await self._update()
                self._device._use()
                return (await self._device.call(self._deviceId,
                                                descriptor.longkey))

            return asyncio.get_event_loop().sync(
                self._raise_on_death(inner), timeout, wait)

        async def _update(self, task=False):
            """assure all properties are properly set

            property settings are cached in self._sethash. This method
            will assure they have properly been sent to the device, or it
            will send it itself. An error is raised if that fails.

            :param task: is an already running update task, if supplied.
            self._last_update_task is checked otherwise.
            """
            if task is False:
                task = self._last_update_task
                self._last_update_task = None
            if task is not None:
                # wait until task is finished
                await task
            if self._sethash:
                # reconfigure the proxy and empty the bulk hash
                sethash, self._sethash = self._sethash, Hash()
                await self._device._sigslot.request(
                    self._deviceId, "slotReconfigure", sethash)

        def _notify_gone(self):
            """The underlying device has gone"""
            self._alive = False
            for task in self._running_tasks:
                task.cancel()
            # actively set the state to UNKNOWN
            h = Hash("state", "UNKNOWN")
            timestamp = get_timestamp()
            timestamp.toHashAttributes(h)
            self._onChanged(h)

        async def _notify_new(self):
            """The underlying device comes back online"""
            schema, _ = await self._device.call(
                self._deviceId, "slotGetSchema", False)
            DeviceClientProxyFactory.updateSchema(self, schema)
            # get configuration
            await self.update_proxy()
            if not self._alive:
                self._alive = True

        async def _raise_on_death(self, func, *args):
            """exec function *func* but raise KaraboError if proxy is orphaned

            This coroutine creates a coroutine from *func*. If the
            device connected to this proxy dies while the coroutine is
            executed, a KaraboError is raised.
            """
            if not self._alive:
                raise KaraboError(
                    f'device "{self._deviceId}" died')
            coro = func(*args)
            task = asyncio.ensure_future(coro)
            self._running_tasks.add(task)
            task.add_done_callback(
                lambda fut: self._running_tasks.discard(fut))
            try:
                return (await task)
            except asyncio.CancelledError:
                if self._alive:
                    raise
                else:
                    raise KaraboError(
                        f'device "{self._deviceId}" died')

        async def initialize_proxy(self):
            """Initialize the proxy by subscribing to updates

            Note: This increments the use counter and will decrease the
            counter after a period to make sure the proxy is connected
            on creation.
            """
            self.__enter__()
            loop = asyncio.get_event_loop()
            loop.call_later(UNSUBSCRIBE_TIMEOUT,
                            WeakMethodRef(self.__exit__), None, None, None)

        def __enter__(self):
            self._used += 1
            if self._used == 1:
                self._device._sigslot.connect(self._deviceId, "signalChanged",
                                              self._device.slotChanged)
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            """Exit the proxy by disconnecting from the signals
            """
            if self._used > 0:
                # Make sure to protect against multiple disconnect calls
                self._used -= 1
                if self._used == 0:
                    self._device._sigslot.disconnect(self._deviceId,
                                                     "signalChanged",
                                                     self._device.slotChanged)

        async def __aenter__(self):
            self._used += 1
            if self._used == 1:
                await self._device._sigslot.async_connect(
                    self._deviceId, "signalChanged", self._device.slotChanged)
            return self

        async def __aexit__(self, exc_type, exc_value, traceback):
            """Exit the proxy by disconnecting from the signals
            """
            if self._used > 0:
                # Make sure to protect against multiple disconnect calls
                self._used -= 1
                if self._used == 0:
                    await self._device._sigslot.async_disconnect(
                        self._deviceId, "signalChanged",
                        self._device.slotChanged)

        def _disconnectSchemaUpdated(self):
            if self._schemaUpdateConnected:
                self._device._sigslot.disconnect(
                    self._deviceId, "signalSchemaUpdated",
                    self._device.slotSchemaUpdated)
            self._schemaUpdateConnected = False

        async def _async_disconnectSchemaUpdated(self):
            if self._schemaUpdateConnected:
                await self._device._sigslot.async_disconnect(
                    self._deviceId, "signalSchemaUpdated",
                    self._device.slotSchemaUpdated)
            self._schemaUpdateConnected = False

        def _disconnect_outputs(self):
            """Disconnect all remote output channels"""
            # Note: In rare cases, e.g. command line, the broker connection
            # is already gone when we are collected, hence we do not throw
            # exceptions from here!
            outputs = list(self._remote_output_channel)
            for channel in outputs:
                channel.disconnect()

        def __del__(self):
            self._disconnect_outputs()
            with suppress(AMQPException):
                self._disconnectSchemaUpdated()
            if self._used > 0:
                self._used = 1
                with suppress(AMQPException):
                    self.__exit__(None, None, None)

        async def delete_proxy(self):
            """Delete the proxy, disconnect all channels and signals"""
            self._disconnect_outputs()
            with suppress(AMQPException):
                await self._async_disconnectSchemaUpdated()
            if self._used > 0:
                self._used = 1
                with suppress(AMQPException):
                    await self.__aexit__(None, None, None)

        async def update_proxy(self):
            """Send out cached changes and get current configuration
            """
            await self._update()
            conf, _ = await self._device.call(self._deviceId,
                                              "slotGetConfiguration")
            self._onChanged(conf)
            return self


class AutoDisconnectProxyFactory(DeviceClientProxyFactory):
    class Proxy(DeviceClientProxyFactory.Proxy):
        def __init__(self, device, deviceId, sync):
            super().__init__(device, deviceId, sync)
            self._interval = 1
            self._lastused = time.time()
            self._task = asyncio.ensure_future(self._connector())

        @synchronous
        def _use(self):
            self._lastused = time.time()
            if self._task.done():
                self._connect()

        @synchronize
        async def _connect(self):
            if self._task.done():
                await self.update_proxy()
                self._task = asyncio.ensure_future(self._connector())

        async def _connector(self):
            with self:
                while True:
                    delta = self._interval - time.time() + self._lastused
                    if delta < 0:
                        return
                    await asyncio.sleep(delta)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import datetime

from sqlalchemy import Column, Text, UniqueConstraint
from sqlmodel import Field, Relationship, SQLModel


class ProjectDomain(SQLModel, table=True):
    __tablename__ = "ProjectDomain"

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(max_length=64, unique=True, nullable=False)
    projects: list["Project"] = Relationship(back_populates="project_domain")


class DatabaseMetadata(SQLModel, table=True):
    __tablename__ = "DatabaseMetadata"

    id: int | None = Field(default=None, primary_key=True)
    schema_version: int = Field(nullable=False, unique=True)


class Project(SQLModel, table=True):
    __tablename__ = "Project"

    id: int | None = Field(default=None, primary_key=True)
    # UUIDs are a legacy from ExistDB. They are being kept and are required
    # for backwards compatibility with ExistDB, and to allow partial
    # migrations from ExistDB (that will be needed if for some reason both
    # storages must coexist for some time; e.g. some topics on ExistDB and some
    # on a Relational Database).
    uuid: str = Field(max_length=64, nullable=False, unique=True)
    name: str = Field(max_length=128, nullable=False)
    is_trashed: bool = Field(default=False)

    date: datetime.datetime | None = Field(default=None, nullable=True)

    last_loaded: datetime.datetime | None = Field(default=None, nullable=True)

    project_domain_id: int = Field(foreign_key="ProjectDomain.id",
                                   nullable=False)
    project_domain: ProjectDomain = Relationship(back_populates="projects")

    device_servers: list["DeviceServer"] = Relationship(
        back_populates="project")

    scenes: list["Scene"] = Relationship(back_populates="project")

    macros: list["Macro"] = Relationship(back_populates="project")

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            f"{self.project_domain.name}/{self.name} ({self.uuid} - "
            f"{self.is_trashed})")


class ProjectSubproject(SQLModel, table=True):
    """The many-to-many relationship between projects and subprojects."""
    __tablename__ = "ProjectSubproject"

    id: int | None = Field(default=None, primary_key=True)
    project_id: int = Field(foreign_key="Project.id", nullable=False)
    subproject_id: int = Field(foreign_key="Project.id", nullable=False)

    # Order of subprojects of a project
    order: int = Field(default=0, nullable=False)

    __table_args__ = (UniqueConstraint("project_id", "subproject_id"), )


class DeviceServer(SQLModel, table=True):
    """A Device Server "captured" from a live topology."""

    __tablename__ = "DeviceServer"

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(max_length=128)
    uuid: str = Field(max_length=64, nullable=False, unique=True)

    date: datetime.datetime | None = Field(default=None, nullable=True)

    project_id: int | None = Field(foreign_key="Project.id", nullable=True)
    project: Project = Relationship(back_populates="device_servers")

    device_instances: list["DeviceInstance"] = Relationship(
        back_populates="device_server")

    # Order of the device servers in a project
    order: int = Field(default=0, nullable=False)


class DeviceInstance(SQLModel, table=True):
    __tablename__ = "DeviceInstance"

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(max_length=256, nullable=False, index=True)
    uuid: str = Field(max_length=64, nullable=False, unique=True)
    # NOTE: if a device instance changes its class_id, all the stored
    #       configurations must be erased upon an user agreement (on normal
    #       Karabo use). For the migration script, an update that changes the
    #       class_id of a device instance that already has some configuration
    #       stored triggers a name mangling and adds a record to the proper
    #       migration transformation registry table.
    class_id: str = Field(max_length=64, nullable=False, index=True)

    device_server_id: int = Field(foreign_key="DeviceServer.id",
                                  nullable=True)

    date: datetime.datetime | None = Field(default=None, nullable=True)

    device_server: DeviceServer = Relationship(
        back_populates="device_instances")
    configs: list["DeviceConfig"] = Relationship(back_populates="device")

    # The order of the devices instances of a device server
    order: int = Field(default=0, nullable=False)


class DeviceConfig(SQLModel, table=True):
    __tablename__ = "DeviceConfig"

    id: int | None = Field(default=None, primary_key=True)
    uuid: str = Field(max_length=64, nullable=False, unique=True)
    name: str = Field(max_length=128, nullable=False)
    # Max size for configuration data is 4Mb
    config_data: str = Field(sa_column=Column(Text(4_194_304), nullable=False))

    date: datetime.datetime | None = Field(default=None, nullable=True)

    device_instance_id: int = Field(foreign_key="DeviceInstance.id",
                                    nullable=True)
    device: DeviceInstance = Relationship(back_populates="configs")

    # The order of the devices configs of a device instance
    order: int = Field(default=0, nullable=False)

    # Is this the active config for among its device instance configs?
    is_active: bool = Field(default=False)

    __table_args__ = (UniqueConstraint("device_instance_id", "name"),)


class Scene(SQLModel, table=True):
    __tablename__ = "Scene"

    id: int | None = Field(default=None, primary_key=True)
    uuid: str = Field(max_length=64, nullable=False, unique=True)
    name: str = Field(max_length=128, nullable=False)
    # Max size for svg_data is 4Mb
    svg_data: str = Field(sa_column=Column(Text(4_194_304), nullable=False))

    date: datetime.datetime | None = Field(default=None, nullable=True)

    project_id: int | None = Field(foreign_key="Project.id", nullable=True)
    project: Project = Relationship(back_populates="scenes")

    # The order of the scenes in a project
    order: int = Field(default=0, nullable=False)


class Macro(SQLModel, table=True):
    __tablename__ = "Macro"

    id: int | None = Field(default=None, primary_key=True)
    uuid: str = Field(max_length=64, nullable=False, unique=True)
    name: str = Field(max_length=128, nullable=False)
    # Max size for macro body is 4Mb
    body: str = Field(sa_column=Column(Text(4_194_304), nullable=False))

    date: datetime.datetime | None = Field(default=None, nullable=True)

    project_id: int | None = Field(foreign_key="Project.id", nullable=True)
    project: Project = Relationship(back_populates="macros")

    # The order of the macros in a project
    order: int = Field(default=0, nullable=False)


class SceneLinkedScene(SQLModel, table=True):
    """The many-to-many relationship between scenes and linked scenes."""
    __tablename__ = "SceneLinkedScene"

    id: int | None = Field(default=None, primary_key=True)
    scene_id: int = Field(foreign_key="Scene.id", nullable=False)
    linked_scene_id: int = Field(foreign_key="Scene.id", nullable=False)

    __table_args__ = (UniqueConstraint("scene_id", "linked_scene_id"), )



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import platform
import re
import socket
from asyncio import (
    CancelledError, Future, IncompleteReadError, Lock, Queue, TimeoutError,
    ensure_future, gather, get_event_loop, open_connection, shield,
    start_server, wait_for)
from collections import deque
from contextlib import closing
from ipaddress import ip_address, ip_network
from struct import calcsize, pack, unpack
from weakref import WeakSet

import numpy
from psutil import net_if_addrs

from karabo.native import (
    AccessLevel, AccessMode, Assignment, Bool, Configurable, Hash, KaraboError,
    MetricPrefix, Node, Schema, String, UInt16, UInt32, Unit, VectorHash,
    VectorRegexString, VectorString, decodeBinary, encodeBinary, get_timestamp,
    isSet)

from .proxy import ProxyBase, ProxyFactory, ProxyNodeBase, SubProxyBase
from .synchronization import (
    KaraboFuture, background, firstCompleted, sleep, synchronize)

DEFAULT_MAX_QUEUE_LENGTH = 2
IPTOS_LOWDELAY = 0x10
RECONNECT_TIMEOUT = 2
SERVER_WAIT_ONLINE = 5

# Simple pattern to detect IP address
IP_REGEX = (r"(?:[0-9]{1,3}\.){3}[0-9]{1,3}"  # IP
            r"(?:\/[0-9]|[1-2][0-9]|3[0-2])?")  # CIDR
IP_PATTERN = re.compile(IP_REGEX)

# Idle time after which keep-alive mechanism
# start checking the connection
TCP_KEEPIDLE = 30
# Interval between probes keep-alive probes
TCP_KEEPINTVL = 5
# Number of ot acknowledged probes after that the
# connection is considered dead
TCP_KEEPCNT = 5


def set_keep_alive(sock: socket.socket):
    """Set the tcp keep alive on a socket `sock`"""
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, TCP_KEEPIDLE)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, TCP_KEEPINTVL)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, TCP_KEEPCNT)


def get_hostname_from_interface(address_range):
    """returns the ip address of the local interfaces

    in case of bad input, it will throw
    """
    # ip_network will throw if address cannot be found
    network = ip_network(address_range)
    for interface in net_if_addrs().values():
        for nic in interface:
            if nic.family != socket.AF_INET:
                continue
            addr = ip_address(nic.address)
            if addr in network:
                return nic.address
    raise ValueError(
        f"{address_range} does not appear to be an IPv4 or IPv6 network")


def get_network_adapter(name):
    addrs = net_if_addrs()
    for addr in addrs.get(name, []):
        if addr.family is socket.AF_INET:
            return addr.address
    raise ValueError(
        f"{name} cannot be found in network configuration")


class CancelQueue(Queue):
    """A queue with an added cancel method

    This uses internal variables from the inherited queue, so it may
    not work when upgrading Python in the future. Maybe we can convince
    the Python guys to add this method to the standard library. Worst case
    we have to copy asyncio's Queue.
    """

    def cancel(self):
        """Cancel all getters and putters currently waiting"""
        for fut in self._putters:
            fut.cancel()
        for fut in self._getters:
            fut.cancel()

    def clear(self):
        self.cancel()
        self._queue.clear()


class RingQueue(CancelQueue):
    def _init(self, maxsize=0):
        """Reimplemented function of `Queue`

        The maxsize is forward from the initializer!
        """
        maxlen = int(maxsize) if maxsize >= 0 else None
        self._queue = deque(maxlen=maxlen)

    def _put(self, item):
        """Reimplemented function of `Queue`"""
        if self.qsize() >= self._maxsize:
            # Declare old task as done!
            self.task_done()
        self._queue.append(item)

    def full(self):
        """Reimplemented function of `Queue`"""
        # XXX: Since we cycle, we should not throw and be full!
        return False


class PipelineMetaData(ProxyBase):
    # Abstract interface configurable
    _allattrs = ["source", "timestamp"]

    source = String(key="source")
    timestamp = Bool(key="timestamp")


class Channel:
    """This class is responsible for reading and writing Hashes to TCP

    It is the low-level implementation of the pipeline protocol."""
    sizeCode = "<I"

    def __init__(self, reader, writer, channelName=None):
        self.reader = reader
        self.writer = writer
        self.drain_lock = Lock()
        self.channelName = channelName

    async def readBytes(self):
        buf = (await self.reader.readexactly(calcsize(self.sizeCode)))
        size, = unpack(self.sizeCode, buf)
        # For large chunks avoid StreamReader.readexactly
        ba = bytearray()
        reader = self.reader
        while len(ba) < size:
            b = await reader.read(size - len(ba))
            if not b:
                raise IncompleteReadError(
                    partial=bytes(ba), expected=size)
            ba.extend(b)
        return ba

    async def readHash(self):
        r = await self.readBytes()
        return decodeBinary(r)

    def writeHash(self, hsh):
        data = encodeBinary(hsh)
        self.writeSize(len(data))
        self.writer.write(data)

    def writeSize(self, size):
        self.writer.write(pack(self.sizeCode, size))

    async def drain(self):
        async with self.drain_lock:
            await self.writer.drain()

    def close(self):
        # WHY IS THE READER NOT CLOSEABLE??
        self.writer.close()

    def sendEndOfStream(self):
        h = Hash("endOfStream", True,
                 "byteSizes", numpy.array([], dtype=numpy.uint32))
        # As the messages transporting data, the eos message has a header
        # (i.e. h) and a body. Now this body is empty, but we still have to
        # send its size in bytes, i.e. zero.
        self.writeHash(h)
        self.writeSize(numpy.uint32(0))

    async def nextChunk(self, chunk_future):
        """write a chunk once availabe and wait for next request

        Wait until the chunk_future becomes available, and write it to the
        output. Then wait for the next request.

        This is all in one method such that we can cancel the future for the
        next chunk if the connection is closed. If this happens, an
        IncompleteReadError is raised which has an empty partial data if the
        channel has been closed properly.
        """
        done, pending, error = await firstCompleted(
            chunk=chunk_future, read=self.readHash(), cancel_pending=False)
        if error:
            for future in pending.values():
                future.cancel()
            raise error.popitem()[1]

        if "chunk" in done:
            chunk = done["chunk"]
            encoded = []
            info = []
            n_chunks = 0
            for data, timestamp in chunk:
                if "endOfStream" in data:
                    # XXX: This is safe as an endOfStream is coming
                    # via an extra method! No chunks are available then!
                    self.sendEndOfStream()
                    continue

                n_chunks += 1
                encoded.append(encodeBinary(data))
                # Create the timestamp information for this packet!
                hsh = Hash("source", self.channelName, "timestamp", True)
                hsh["timestamp", ...] = timestamp.toDict()
                info.append(hsh)

            if n_chunks:
                nData = numpy.uint32(n_chunks)
                sizes = numpy.array([len(d) for d in encoded],
                                    dtype=numpy.uint32)
                h = Hash("nData", nData,
                         "byteSizes", sizes,
                         "sourceInfo", info)
                # For future: Replace bytesizes with `bufferSetLayout` and
                # special content to enable copyless deserialization
                self.writeHash(h)
                self.writeSize(sizes.sum())
                self.writer.writelines(encoded)

        if "read" in pending:
            # NOTE: We accept the fact that also ``read`` can be in ``done``!
            # If ``read`` is still ``pending``, we await it here!
            message = await pending["read"]
            assert message["reason"] == "update"
        else:
            message = done["read"]
            assert message["reason"] == "update"
            text = ("{} - Received chunk acknowledgement ``read``. A chunk "
                    "future was retrieved before: {}".format(self.channelName,
                                                             "chunk" in done))
            print(text)
            await sleep(0)


class NetworkInput(Configurable):
    """The input channel

    This represents an input channel. It is typically not used directly,
    instead you should just declare a :cls:`InputChannel`.
    """
    displayType = 'InputChannel'
    classId = 'InputChannel'

    # Internal name to be set by the input channel
    _name = None

    missingConnections = VectorString(
        displayedName="Missing Connections",
        description="Output channels from 'Configured Connections' "
                    "that are not connected",
        defaultValue=[],
        accessMode=AccessMode.READONLY)

    @VectorRegexString(
        displayedName="Configured Connections",
        description="A list of output channels to receive data from, format: "
                    "<instance ID>:<channel name>",
        assignment=Assignment.OPTIONAL, accessMode=AccessMode.INITONLY,
        regex=r"^[a-zA-Z0-9\.\-\:\/\_]+$",
        defaultValue=[])
    async def connectedOutputChannels(self, value):
        # Basic name protection is implemented here, as it can cause hickups!
        outputs = {output.strip() for output in value}
        if not outputs:
            self.connectedOutputChannels = []
            return

        close = self.connected.keys() - outputs
        for k in close:
            self.connected[k].cancel()
        for output in outputs:
            await self.connectChannel(output)

    async def connectChannel(self, channel):
        """Connect to a single Outputchannel
        """
        if channel in self.connected and not self.connected[channel].done():
            return
        self._update_missing_connections(channel, missing=True)
        # Make sure to always set the instance correctly!
        loop = get_event_loop()
        task = loop.create_task(self.start_channel(channel),
                                instance=self.parent.get_root())
        self.connected[channel] = task
        self.connectedOutputChannels = list(self.connected)

    dataDistribution = String(
        displayedName="Data Distribution",
        description="The way data is fetched from the connected output "
                    "channels (shared/copy)",
        options=["shared", "copy"], assignment=Assignment.OPTIONAL,
        defaultValue="copy", accessMode=AccessMode.RECONFIGURABLE)

    onSlowness = String(
        displayedName="On Slowness",
        description="Policy for what to do if this input is too slow for the "
                    "fed data rate (only used in copy mode)",
        options=["queueDrop", "drop", "wait"],
        assignment=Assignment.OPTIONAL, defaultValue="drop",
        accessMode=AccessMode.RECONFIGURABLE)

    maxQueueLength = UInt32(
        defaultValue=DEFAULT_MAX_QUEUE_LENGTH,
        minInc=1,
        displayedName="Max. Queue Length Output Channels",
        description="Maximum number of data items to be queued by connected "
                    "Output Channels (only in copy mode and for "
                    "queueDrop policy - Output may force a stricter limit)",
        accessMode=AccessMode.INITONLY)

    delayOnInput = UInt32(
        defaultValue=0,
        displayedName="Delay on Input channel",
        description="Some delay before informing output channel about "
                    "readiness for next data.",
        unitSymbol=Unit.SECOND,
        metricPrefixSymbol=MetricPrefix.MILLI,
        accessMode=AccessMode.INITONLY)

    def __init__(self, config):
        super().__init__(config)
        self.connected = {}
        self.handler_lock = Lock()

    async def connect_handler(self, output):
        # XXX: Please keep this for the time being.
        print(f"NetworkInput connect handler called by {output}!")

    async def close_handler(self, output):
        # XXX: Please keep this for the time being.
        print(f"NetworkInput close handler called by {output}!")

    async def end_of_stream_handler(self, cls):
        # XXX: Please keep this for the time being.
        print(f"EndOfStream handler called by {cls}!")

    async def call_handler(self, func, *args):
        """Call a network input handler under mutex and error protection"""
        async with self.handler_lock:
            try:
                await wait_for(get_event_loop().run_coroutine_or_thread(
                    func, *args), timeout=5)
            except CancelledError:
                # Cancelled Error is handled in the `start_channel` method
                pass
            except TimeoutError:
                self.device_logger.error(
                    f"timeout in handler {func.__name__} in stream")
            except Exception:
                self.device_logger.exception("error in stream")

    @property
    def device_logger(self):
        """Convenience method to retrieve device logger"""
        return self.parent.get_root().logger

    async def start_channel(self, output):
        """Connect to the output channel with Id 'output' """
        try:
            root = self.parent.get_root()
            logger = root.logger
            logger.info(f"Trying to connect to channel {output}")
            instance, name = output.split(":")
            h = await root._call_once_alive(
                instance, "slotGetOutputChannelInformation",
                Hash("channelId", name))
            ok = h["success"]
            info = h["info"]
            if not ok:
                logger.info(
                    f"Connecting to channel that was not there {output}")
                # Start fresh!
                await sleep(RECONNECT_TIMEOUT)
                return (await self.start_channel(output))

            if self.raw:
                cls = None
            else:
                # schema from output channel we are talking with
                schema = info.get("schema", None)
                if schema is None:
                    schema, _ = await root.call(
                        instance, "slotGetSchema", False)
                cls = ProxyFactory.createProxy(
                    Schema(name=name, hash=schema.hash[name]["schema"]))

            reader, writer = await open_connection(
                info["hostname"], int(info["port"]), limit=2 ** 22)

            # Inform about exiting connection
            self._update_missing_connections(output, missing=False)
            sock = writer.get_extra_info("socket")
            assert sock.getsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY)
            # Set low delay according to unix manual IPTOS_LOWDELAY 0x10
            sock.setsockopt(socket.SOL_IP, socket.IP_TOS, IPTOS_LOWDELAY)
            if platform.system() == "Linux":
                set_keep_alive(sock)
            channel = Channel(reader, writer, channelName=output)
            with closing(channel):
                # Tell the world we are connected before we start
                # requesting data
                await shield(self.call_handler(self.connect_handler, output))
                instance_id = f"{root.deviceId}:{self._name}"
                cmd = Hash("reason", "hello",
                           "instanceId", instance_id,
                           "memoryLocation", "remote",
                           "dataDistribution", self.dataDistribution,
                           "onSlowness", self.onSlowness,
                           "maxQueueLength", self.maxQueueLength.value)
                channel.writeHash(cmd)
                while (await self.processChunk(channel, cls, instance_id)):
                    await sleep(self.delayOnInput)
                else:
                    # We still inform when the connection has been closed!
                    self._update_missing_connections(output, missing=True)
                    await shield(self.call_handler(self.close_handler, output))
                    # Start a new roundtrip, but don't create a new task!
                    await sleep(RECONNECT_TIMEOUT)
                    await self.start_channel(output)
        except CancelledError:
            # Note: Happens when we are destroyed or disconnected!
            await shield(self.call_handler(self.close_handler, output))
            self._update_configured_connections(output)
        except Exception:
            logger = self.device_logger
            logger.info(f"Channel `{output}` did not close gracefully.")
            self._update_missing_connections(output, missing=True)
            await shield(self.call_handler(self.close_handler, output))
            # We might get cancelled during sleeping!
            try:
                await sleep(RECONNECT_TIMEOUT)
                await self.start_channel(output)
            except CancelledError:
                self._update_configured_connections(output)
                # Already missing output, no need to update missing

    def _update_configured_connections(self, output):
        """Update and remove channel `output` from connected"""
        # Graceful, due to cancellation it might get out of sync
        self.connected.pop(output, None)
        self.connectedOutputChannels = list(self.connected)

    def _update_missing_connections(self, output, missing):
        """Update the missing connections of the `InputChannel`

        :param missing: Boolean if the connection is missing.
        """
        connections = set(self.missingConnections.value)
        if missing:
            connections.add(output)
        else:
            connections.discard(output)
        self.missingConnections = list(connections)

    async def processChunk(self, channel, cls, output_id=""):
        try:
            header = await channel.readHash()
        except IncompleteReadError as e:
            if e.partial:
                raise
            else:
                root = self.parent.get_root()
                root.logger.info("stream %s finished", channel.channelName)
                return False
        data = await channel.readBytes()
        # `readHash` must happen first, but then we can already
        # request for more updates before processing!
        cmd = Hash("reason", "update",
                   "instanceId", output_id)
        channel.writeHash(cmd)
        if "endOfStream" in header:
            await shield(self.call_handler(
                self.end_of_stream_handler, channel.channelName))
            return True
        pos = 0
        for length, meta_hash in zip(header["byteSizes"],
                                     header["sourceInfo"]):
            chunk = decodeBinary(data[pos:pos + length])
            meta = PipelineMetaData()
            meta._onChanged(meta_hash)
            if self.raw:
                await shield(self.call_handler(self.handler, chunk, meta))
            else:
                proxy = cls()
                proxy._onChanged(chunk)
                await shield(self.call_handler(self.handler, proxy, meta))
            pos += length
        return True


class PipelineContext(NetworkInput):
    """This represents a context specifc input channel connection to a karabo
        device. The context is not connected automatically, but may be
        connected using :meth:`async with` or :meth:`with`::

            channel = PipelineContext("deviceId:output")
            async with channel:
                data, meta = await channel.get_data()

            with channel:
                await channel.get_data()

        It is possible to ask for the connection of the context using::

            async with channel:
                if not channel.is_alive():
                    await channel.wait_connected()
    """

    def __init__(self, output, configuration={}):
        configuration.update(dict(dataDistribution="copy", onSlowness="drop"))
        super().__init__(configuration)
        self.raw = True
        self.parent = self

        self._task = None
        self._output = output
        self._initialized = False

        self._connected = False
        self._queue = CancelQueue()
        self._connect_futures = set()

    # Public interface
    # ----------------------------------------------------------------------

    @synchronize
    async def wait_connected(self):
        """Wait for the pipeline context to be connect

        with PipelineContext(output) as ctx:
            await ctx.wait_connected()

            # do something ...
        """
        if self._connected:
            return True

        future = Future()
        self._connect_futures.add(future)
        return await future

    def is_alive(self):
        """Public method to provide information if we are connected and alive
        """
        return self._connected

    def __enter__(self):
        self._task = background(self._connect())
        return self

    def __exit__(self, exc, value, tb):
        self._disconnect()

    def size(self):
        return self._queue.qsize()

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, exc, value, tb):
        return self.__exit__(exc, value, tb)

    @synchronize
    async def get_data(self):
        return await self._queue.get()

    # Private interface
    # ----------------------------------------------------------------------

    def get_root(self):
        """Reimplemented function of `Configurable`"""
        return get_event_loop().instance()

    async def connect_handler(self, output):
        """Reimplemented function of `NetworkInput`"""
        self._connected = True
        for future in self._connect_futures:
            future.set_result(True)

        self._connect_futures = set()

    async def close_handler(self, output):
        """Reimplemented function of `NetworkInput`"""
        self._connected = False

    def handler(self, data, meta):
        """Reimplemented function of `NetworkInput`"""
        self._queue.put_nowait((data, meta))

    async def _connect(self):
        output = self._output
        # Similar to the `OutputProxy`, we _run the Configurable and
        # attach the task.
        try:
            if not self._initialized:
                await self._run()
                self._initialized = True
            self.connected[output] = self._task
            await self.start_channel(output)
        finally:
            self._task = None

    def _disconnect(self):
        """Disconnect from the output channel"""
        if self._task is not None and not self._task.done():
            self._task.cancel()
        self._connected = False
        self._queue.clear()


class InputChannel(Node):
    """Declare an input channel in a device

    This descriptor declares an input channel of a devices. It is most
    conveniently used as a decorator to the handler for the incoming data::

       @InputChannel()
       def input(self, data, meta):
           # do something useful with data

    The handler is passed the sent data as well as a meta object which has
    two attributes: ``meta.source`` is the name of the connected output channel
    (device id and channel name join with a colon), while ``meta.timestamp`` is
    a boolean which is always `True`, but has the timestamp of the arriving
    data.

    Should it be necessary to act upon a channel closing, a close handler may
    be installed::

        @input.close
        def input(self, name):
            # close the input channel `name`

    Should it be necessary to act upon endOfStream, a handler may be
    installed::

        @input.endOfStream
        def input(self, name):
            # EndOfStream handler has been called by `name`

    Should it be necessary to act upon `connect`, a handler may be
    installed::

        @input.connect
        def input(self, name):
            # Connect handler has been called by `name`

    :param raw: especially if the sender has not declared the schema of the
        data it sends, one can set `raw` to `True` in order to get the
        uninterpreted hash. The `data` object will be empty if the sender has
        not declared a schema.
    """
    connect_handler = None
    close_handler = None
    end_of_stream_handler = None

    def __init__(self, raw=False, **kwargs):
        super().__init__(NetworkInput, **kwargs)
        self.raw = raw

    def _initialize(self, instance, value):
        """This method is called on initialization

        Called via Configurable checkedInit for every descriptor.

        The `value` still is the bare Hash value, as it came from the network!
        """
        ret = super()._initialize(instance, value)
        channel = instance.__dict__[self.key]
        channel.raw = self.raw
        channel.handler = self.handler.__get__(instance, type(instance))
        channel.parent = instance
        channel._name = self.key
        if self.connect_handler is not None:
            channel.connect_handler = self.connect_handler.__get__(
                instance, type(instance))
        if self.close_handler is not None:
            channel.close_handler = self.close_handler.__get__(
                instance, type(instance))
        if self.end_of_stream_handler is not None:
            channel.end_of_stream_handler = self.end_of_stream_handler.__get__(
                instance, type(instance))
        return ret

    def connect(self, func):
        self.connect_handler = func
        return self

    def close(self, func):
        self.close_handler = func
        return self

    def endOfStream(self, func):
        self.end_of_stream_handler = func
        return self

    def __call__(self, handler):
        if self.description is None:
            self.description = handler.__doc__
        self.handler = handler
        return self


class OutputProxyNode(ProxyNodeBase):
    """Descriptor representing a Proxy for an output channel"""

    def __init__(self, key, node, prefix, **kwargs):
        self.longkey = f"{prefix}{key}."
        sub = ProxyFactory.createNamespace(node, self.longkey)
        cls = type(key, (OutputProxy,), sub)
        super().__init__(key=key, cls=cls, **kwargs)

    def __get__(self, instance, owner):
        ret = super().__get__(instance, owner)
        ret.longkey = self.longkey
        return ret


ProxyFactory.node_factories["OutputChannel"] = OutputProxyNode


class OutputProxy(SubProxyBase):
    """A Proxy for an output channel

    This represents an output channel on a device. Output channels are
    not connected automatically, but may be connected using :meth:`connect`::

        device = connectDevice("someDevice")
        # Assuming the output channel has the key `output`
        device.output.connect()

    The default pipeline settings:

        dataDistribution = "copy"
        onSlowness = "drop"

    By default the proxy applies the meta data timestamp to the received data.
    This can be switched to the Hash timestamp by setting the ``meta``
    attribute to ``False``.

        device = connectDevice("someDevice")
        device.output.meta = False
        device.output.connect()
    """
    schema = None

    def __init__(self):
        self.task = None
        self.initialized = False
        # Define if we apply meta data to the pipeline data on the proxy
        self.meta = True
        self._data_handler = None
        self._eos_handler = None
        self._connect_handler = None
        self._close_handler = None
        self.networkInput = None

    def setDataHandler(self, handler):
        """Redirect the output of the pipelining proxy before connecting

        NOTE: The handler must take two arguments ``data`` and ``meta``.
        """
        if self.task is not None:
            raise RuntimeError("Setting a data handler must happen before "
                               "connecting to the output channel!")
        self._data_handler = handler

    def setEndOfStreamHandler(self, handler):
        """Redirect the endOfStream of the pipelining proxy before connecting

        NOTE: The handler must take one argument ``channelname``.
        """
        if self.task is not None:
            raise RuntimeError("Setting an endOfStream handler must happen "
                               "before connecting to the output channel!")
        self._eos_handler = handler

    def setConnectHandler(self, handler):
        """Redirect the connect signal of the pipelining proxy

        NOTE: The handler must take one argument ``channelname``.
        """
        if self.task is not None:
            raise RuntimeError("Setting a connected handler must happen "
                               "before connecting to the output channel!")
        self._connect_handler = handler

    def setCloseHandler(self, handler):
        """Redirect the closing signal of the pipelining proxy

        NOTE: The handler must take one argument ``channelname``.
        """
        if self.task is not None:
            raise RuntimeError("Setting a close handler must happen "
                               "before connecting to the output channel!")
        self._close_handler = handler

    async def handler(self, data, meta):
        # Only apply if we have a schema!
        if self.schema is not None:
            if self.meta and meta.timestamp:
                self._parent._onChanged_timestamp_r(
                    data, meta.timestamp.timestamp, self.schema)
            else:
                self._parent._onChanged_r(data, self.schema)

    def connect(self):
        """Connect to the output channel

        The connection is always a copy/drop connection, meaning only samples
        of the data reach this proxy, without influencing other consumer.
        For a fully featured input write a device with a
        :class:`InputChannel`.
        """
        if self.task is not None:
            return

        self.networkInput = NetworkInput(dict(
            dataDistribution="copy", onSlowness="drop"))
        self.networkInput.raw = True
        if self._data_handler is not None:
            self.networkInput.handler = self._data_handler
        else:
            self.networkInput.handler = self.handler
        if self._eos_handler is not None:
            self.networkInput.end_of_stream_handler = self._eos_handler
        if self._connect_handler is not None:
            self.networkInput.connect_handler = self._connect_handler
        if self._close_handler is not None:
            self.networkInput.close_handler = self._close_handler
        output = ":".join((self._parent.deviceId, self.longkey))
        # Add our channel to the proxy tracking
        self._parent._remote_output_channel.add(self)
        self.networkInput.parent = self._parent._device
        # Track task correctly according to the Eventloop
        loop = get_event_loop()
        if not loop.sync_set:
            self.task = loop.create_task(self._connect(output),
                                         instance=self._parent._device)
        else:
            self.task = background(self._connect(output))

    async def _connect(self, output):
        try:
            if not self.initialized:
                await self.networkInput._run()
                self.initialized = True
            self.networkInput.connected[output] = self.task
            await self.networkInput.start_channel(output)
        finally:
            self.task = None

    def disconnect(self):
        """Disconnect from the output channel"""
        if self in self._parent._remote_output_channel:
            self._parent._remote_output_channel.remove(self)
        if self.task is not None:
            if isinstance(self.task, KaraboFuture):
                ensure_future(self.task.cancel())
            else:
                self.task.cancel()
            self.task = None


class ConnectionTable(Configurable):
    remoteId = String(
        displayedName="Remote ID",
        description="Id of remote input channel")

    dataDistribution = String(
        displayedName="Distribution",
        description="Data distribution of the input channel: shared or copy")

    onSlowness = String(
        displayedName="On slowness",
        description="Data handling policy in case of slowness if data "
                    "distribution is copy: drop, wait, queueDrop")

    memoryLocation = String(
        displayedName="MemoryLocation",
        description="Cache Memory class location: can be remote or local")

    remoteAddress = String(
        displayedName="Remote IP",
        description="Remote TCP address of active connection")

    remotePort = UInt16(
        displayedName="Remote Port",
        description="Remote port of active connection")

    localAddress = String(
        displayedName="Local IP",
        description="Local TCP address of active connection")

    localPort = UInt16(
        displayedName="Local Port",
        description="Local port of active connection")


class NetworkOutput(Configurable):
    displayType = 'OutputChannel'
    classId = 'OutputChannel'

    port = UInt32(
        displayedName="Port",
        description="Port number for TCP connection",
        requiredAccessLevel=AccessLevel.EXPERT,
        assignment=Assignment.OPTIONAL,
        defaultValue=0,
        accessMode=AccessMode.INITONLY)

    noInputShared = String(
        displayedName="No Input (Shared)",
        description="What to do if currently no share-input channel is "
                    "available for writing to",
        options=["drop", "wait", "queueDrop"],
        assignment=Assignment.OPTIONAL, defaultValue="drop",
        accessMode=AccessMode.INITONLY)

    @String(
        displayedName="Hostname",
        description="The hostname which connecting clients will be routed to. "
                    "Network ranges in the CIDR notation are accepted, "
                    "e.g. 0.0.0.0/24",
        assignment=Assignment.OPTIONAL, defaultValue="default",
        accessMode=AccessMode.INITONLY)
    async def hostname(self, value):
        if value == "default":
            hostname = socket.gethostname()
        elif IP_PATTERN.match(value):
            hostname = get_hostname_from_interface(value)
        else:
            hostname = get_network_adapter(value)

        instance = get_event_loop().instance()

        def serve(reader, writer):
            """Create a serve task on the eventloop to track instance

               We need to pass the instance in order to cancel all the tasks
               related to that instance once the instance dies.
            """
            loop = get_event_loop()
            channel_task = loop.create_task(
                self.serve(reader, writer), instance=instance)
            self.active_channels.add(channel_task)
            if not self.alive:
                ts = get_timestamp().toLocal()
                print(f"{ts}: Cancelling channel, going down as instructed.")
                loop.call_soon_threadsafe(channel_task.cancel)

        port = int(self.port) if isSet(self.port) else 0
        self.server = await start_server(serve, host=hostname,
                                         port=port)
        self.hostname = value
        self.address = hostname

    address = String(
        displayedName="Address",
        description="The address resolved from the hostname",
        accessMode=AccessMode.READONLY)

    connections = VectorHash(
        rows=ConnectionTable,
        defaultValue=[],
        displayedName="Connections",
        description="Table of active connections",
        accessMode=AccessMode.READONLY)

    maxQueueLength = UInt32(
        defaultValue=DEFAULT_MAX_QUEUE_LENGTH,
        minInc=2,  # 1 would be equivalent to 'drop' policy
        displayedName="Max. Queue Length",
        description="Maximum queue length accepted from 'copy' mode input "
                    "channels with 'queueDrop' policy",
        accessMode=AccessMode.INITONLY)

    def __init__(self, config):
        super().__init__(config)
        self.copy_queues = []
        self.wait_queues = []
        self.copy_futures = []
        self.active_channels = WeakSet()
        self.has_shared = 0
        self.channelName = ""
        self.shared_queue = CancelQueue(0 if self.noInputShared == "queueDrop"
                                        else 1)
        self.server = None
        # By default, an output is considered alive until closed
        self.alive = True

    def is_serving(self):
        """Return if the output channel can serve data"""
        return self.alive and self.server is not None

    async def wait_server_online(self):
        """Wait until the serving tcp server comes online"""
        total_time = SERVER_WAIT_ONLINE
        interval_time = 0.05  # [s]
        while self.server is None:
            await sleep(interval_time)
            total_time -= interval_time
            if total_time <= 0:
                break
        if self.server is None:
            raise KaraboError("Tcp server of output should be available ...")

    def getInformation(self, channelName):
        self.channelName = channelName
        if self.server is None:
            raise KaraboError(
                f"Trying to get output information for {channelName}, "
                "but output server not available.")
        host, port = self.server.sockets[0].getsockname()
        return Hash("connectionType", "tcp", "hostname", host,
                    "port", numpy.uint32(port))

    async def serve(self, reader, writer):
        channel = Channel(reader, writer, self.channelName)
        distribution = "unknown"
        remote_host = "unknown"
        remote_port = 0
        try:
            message = await channel.readHash()
            assert message["reason"] == "hello"

            # Start parsing the hello message!
            channel_name = message["instanceId"]
            distribution = message["dataDistribution"]
            slowness = message["onSlowness"]
            maxQueueLength = int(message.get("maxQueueLength",
                                             DEFAULT_MAX_QUEUE_LENGTH))
            # Protect against unnecessary high values from outside!
            # The MDL does not have a memory queue ...
            maxQueueLength = min(maxQueueLength,
                                 int(self.maxQueueLength.value))
            local_host, local_port = writer.get_extra_info("sockname")
            remote_host, remote_port = writer.get_extra_info("peername")
            # Set the connection table entry
            entry = (channel_name, distribution, slowness, "remote",
                     remote_host, remote_port, local_host, local_port)
            self.connections.extend(entry)

            if distribution == "shared":
                self.has_shared += 1
                while True:
                    await channel.nextChunk(self.shared_queue.get())
            elif slowness == "drop":
                while True:
                    future = Future()
                    self.copy_futures.append(future)
                    await channel.nextChunk(future)
            else:
                if slowness == "queueDrop":
                    queue = RingQueue(maxQueueLength)
                else:
                    queue = CancelQueue(1)
                if slowness == "wait":
                    queues = self.wait_queues
                else:
                    queues = self.copy_queues
                queues.append(queue)
                try:
                    while True:
                        await channel.nextChunk(queue.get())
                finally:
                    queues.remove(queue)
                    queue.cancel()
        except CancelledError:
            # If we are cancelled, we should close ourselves
            # Note: This happens when the output channel is closed and
            # when the device is destroyed.
            await self.close()
        except IncompleteReadError as e:
            if e.partial:  # if the input got properly closed, partial is empty
                raise
        finally:
            # XXX: Rewrite to channel name!
            if distribution == "shared":
                self.has_shared -= 1
                if self.has_shared == 0:
                    # XXX: Make sure nothing is stuck and cancel all futures
                    # where the consumer does not shutdown gracefully!
                    self.shared_queue.cancel()
            for index, row in enumerate(self.connections.value):
                if (row['remoteAddress'] == remote_host and
                        row['remotePort'] == remote_port):
                    del self.connections[index]
            channel.close()

    def writeChunkNoWait(self, chunk):
        if (self.has_shared > 0 and self.noInputShared != "wait"
                and not self.shared_queue.full()):
            self.shared_queue.put_nowait(chunk)
        for future in self.copy_futures:
            if not future.done():
                future.set_result(chunk)
        self.copy_futures = []
        for queue in self.copy_queues:
            queue.put_nowait(chunk)

    async def writeChunk(self, chunk):
        tasks = [sleep(0)]
        try:
            self.writeChunkNoWait(chunk)
            if self.has_shared > 0 and self.noInputShared == "wait":
                tasks.append(self.shared_queue.put(chunk))
            for queue in self.wait_queues:
                tasks.append(queue.put(chunk))
        finally:
            await gather(*tasks, return_exceptions=True)

    async def writeEndOfStream(self):
        """Send an endOfStream to the clients"""
        timestamp = get_timestamp()
        await self.writeChunk([(Hash("endOfStream", True), timestamp)])

    async def writeData(self, timestamp=None):
        """Send the applied values with given output schema to the clients

        Requires an output channel with a schema
        """
        hsh = self.schema.configurationAsHash()
        if timestamp is None:
            timestamp = get_timestamp()
        await self.writeChunk([(hsh, timestamp)])

    async def writeRawData(self, hsh, timestamp=None):
        """Send raw hash data via the output channel to the clients

        This method can be used if the output channel is used in 'raw' mode,
        e.g. does not have a schema.
        """
        assert isinstance(hsh, Hash)

        if timestamp is None:
            timestamp = get_timestamp()
        await self.writeChunk([(hsh, timestamp)])

    def writeDataNoWait(self, timestamp=None):
        hsh = self.schema.configurationAsHash()
        if timestamp is None:
            timestamp = get_timestamp()
        self.writeChunkNoWait([(hsh, timestamp)])

    def writeRawDataNoWait(self, hsh, timestamp=None):
        assert isinstance(hsh, Hash)

        if timestamp is None:
            timestamp = get_timestamp()
        self.writeChunkNoWait([(hsh, timestamp)])

    def setChildValue(self, key, value, descriptor):
        """Set the child values on the Configurable

        Overwrite the method of Configurable to prevent sending values for
        the pipeline data schema!
        """
        if "." in key:
            return
        super().setChildValue(key, value, descriptor)

    async def close(self):
        """ Cancel all channels and close the server sockets"""
        if not self.alive:
            return

        self.alive = False
        await self.wait_server_online()
        # Perform cleanup only after server is online
        for channel_task in list(self.active_channels):
            channel_task.cancel()
            await channel_task

        self.active_channels = WeakSet()
        self.server.close()
        server = self.server
        self.server = None
        await server.wait_closed()


class SchemaNode(Node):
    def toDataAndAttrs(self, instance):
        """The Output Channel Schema is not allowed to provide data"""
        return Hash(), {}


class OutputChannel(Node):

    def __init__(self, cls=None, **kwargs):
        if cls is None:
            # Child check for output schema
            Output = NetworkOutput
        else:
            assert issubclass(cls, Configurable)
            cls.displayType = "OutputSchema"

            class Output(NetworkOutput):
                schema = SchemaNode(cls)

        super().__init__(Output, **kwargs)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import os.path as op
import re
import subprocess
from functools import partial
from os import environ

from packaging.version import InvalidVersion, Version

# NOTE: Only stdlib imports are allowed here! This module is used when
# installing the karabo package.


class SvnParseError(ValueError):
    pass


def _get_development_version(path):
    """ Return the current development version.
    """
    commit_count = '0'
    generators = [partial(f, path)
                  for f in (svn_version, jsvn_version, git_version)]
    for gen in generators:
        vers_dict = gen()
        if vers_dict['vcs_revision'] != 'Unknown':
            commit_count = vers_dict['revision_count']
            break

    return commit_count


def _get_karabo_framework_path():
    """ Get the filesystem location of the running Karabo framework.
    """
    karabo = environ.get('KARABO')
    if karabo is None:
        msg = ("$KARABO env variable not defined! Source the activate script "
               "for the desired Karabo framework to address this issue.")
        raise RuntimeError(msg)

    return karabo


def _parse_svn_part(info_str, part_name):
    """ Extract various information from the output of `svn info`

    part_name = 'revision': Gives the current SVN revision number
    part_name = 'branch': Gives the SVN branch name
    """
    svn_parts = {
        'revision': r'Revision\: (\d+)',
        'branch': r'URL\: .+/(.+)',
    }
    match = re.search(svn_parts[part_name], info_str)
    if match is not None:
        return match.groups()[0]
    raise SvnParseError(f'SVN {part_name} not found!')


def get_karabo_framework_version(strict_fallback=True) -> Version:
    """ Returns the current Karabo version contained in the Framework VERSION
    file.

    Parameter:
    strict_fallback(bool): when True, considers only the "major.minor.patch"
    portion as a fallback strategy in case an invalid version string is read
    from the VERSION file.

    Returns:
    A packaging.version.Version with the value stored in the VERSION file if
    that value (or its "major.minor.patch" portion) is a valid version string.

    Raises:
    A packaging.version.InvalidVersion if the value in the VERSION file is
    an invalid version string even after the fallback strategy is applied.
    """
    str_version = _read_version_file()
    version = None
    try:
        version = Version(str_version)
    except InvalidVersion:
        if strict_fallback:
            str_version = _filter_major_minor_patch(str_version)
            version = Version(str_version)
        else:
            raise
    return version


def _read_version_file() -> str:
    karabo_path = _get_karabo_framework_path()
    version_path = op.join(karabo_path, 'VERSION')
    with open(version_path) as fp:
        version = fp.read().strip()
    return version


def _filter_major_minor_patch(version_str: str) -> str:
    pattern = r'(?P<major>\d+)\.(?P<minor>\d+)\.?(?P<patch>\d*)?'
    matches = re.match(pattern, version_str)
    return matches[0] if matches else ""


def get_karabo_version():
    """Return the current Karabo Python package version."""
    try:
        from karabo._version import full_version
    except ImportError:
        from karabo._version import version
        from karabo.common.packaging import utils
        full_version = utils.extract_full_version(version)
    return full_version


def get_package_version(path):
    """ Get the package version for a device's repository.

    Returns a PEP 440 compatible version string which is a combination of the
    active Karabo framework's version and the commit count of the device's
    repository.
    """
    path = op.normpath(path)
    karabo_version = get_karabo_version()
    dev_version = _get_development_version(path)
    extra = f'.dev{dev_version}' if dev_version != '0' else ''

    return karabo_version + extra


def git_version(path):
    """ Return the git revision as a string and a commit count
    """
    try:
        # Find a tag matching the glob *.*.* (version number prefixed tags)
        cmd = 'git describe --match *.*.*'.split(' ')
        out = subprocess.check_output(cmd, cwd=path).decode('ascii')
    except (subprocess.CalledProcessError, OSError):
        out = ''

    # Git will return something like NN.NN.NN[(a|b|rc)NN]?-NN-XXXXXX
    # The group names in this regex explain the parts
    # The only required part of the regex is the NN.NN.NN. All others optional.
    expr = (
        r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<micro>\d+)(?P<abrc>(a|b|rc)\d+)?'
        r'(\-)?(?P<count>\d+)?(-g)?(?P<hash>[a-fA-F0-9]+)?'
    )
    match = re.match(expr, out)
    if match is None:
        revision, count = 'Unknown', '0'
        version, released = '0.0.0', False
    else:
        revision = match.group('hash') or ''
        count = match.group('count') or '0'
        version = '.'.join([match.group(n)
                            for n in ('major', 'minor', 'micro')])
        released = (revision == '' and count == '0')

    return {
        'vcs_revision': revision,
        'revision_count': count,
        'version': version,
        'released': released,
    }


def svn_version(path, svn_cmd='svn'):
    """ Return the SVN revision as a string and a commit count.
    """
    cmd = [svn_cmd, 'info', path]
    try:
        out = subprocess.check_output(cmd).decode('ascii')
        svn_branch = _parse_svn_part(out, 'branch')
        svn_count = _parse_svn_part(out, 'revision')
    except (subprocess.CalledProcessError, OSError, SvnParseError):
        return {'vcs_revision': 'Unknown', 'revision_count': '0',
                'version': '0.0.0', 'released': False}

    return {
        'vcs_revision': f'{svn_branch}@r{svn_count}',
        'revision_count': svn_count,
        'version': svn_branch,
        'released': svn_count == '0',
    }


# Define this where it can be imported.
jsvn_version = partial(svn_version, svn_cmd='jsvn')


class device_scm_version:
    """Backward compatible karabo device tags compliant to PEP440"""

    def __init__(self, root_path, file_path):
        self.root = root_path
        self.write_to = file_path

    def __call__(self):
        """setuptools_scm allows a callable to be specified
        in the setup function"""
        return {
            "root": self.root,
            "write_to": self.write_to,
        }
